# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations
from collections.abc import Container
import ctypes
from typing import (
    List, NamedTuple, Sequence, Set, Union, Optional, cast, TYPE_CHECKING)

import numpy

from spinn_utilities.abstract_base import abstractmethod
from spinn_utilities.overrides import overrides
from spinn_utilities.ranged.abstract_sized import Selector

from pacman.model.graphs import AbstractVertex
from pacman.model.graphs.common import Slice
from pacman.model.placements import Placement
from pacman.utilities.utility_calls import get_keys

from spinn_front_end_common.interface.ds import (
    DataSpecificationBase, DataSpecificationGenerator,
    DataSpecificationReloader)
from spinn_front_end_common.interface.provenance import ProvenanceWriter

from spynnaker.pyNN.data import SpynnakerDataView
from spynnaker.pyNN.utilities.utility_calls import get_n_bits
from spynnaker.pyNN.models.abstract_models import AbstractNeuronExpandable
from spynnaker.pyNN.models.current_sources import CurrentSourceIDs
from spynnaker.pyNN.utilities.utility_calls import convert_to
if TYPE_CHECKING:
    from spynnaker.pyNN.models.neuron import PopulationVertex
    from spynnaker.pyNN.models.neuron.neuron_data import NeuronData
    from spynnaker.pyNN.models.current_sources import AbstractCurrentSource


class NeuronProvenance(ctypes.LittleEndianStructure):
    """
    Provenance items from neuron processing.
    """
    _fields_ = [
        # The timer tick at the end of simulation
        ("current_timer_tick", ctypes.c_uint32),
        # The number of misses of TDMA time slots
        ("n_tdma_misses", ctypes.c_uint32),
        # The earliest send time within any time step
        ("earliest_send", ctypes.c_uint32),
        # The latest send time within any time step
        ("latest_send", ctypes.c_uint32)
    ]

    N_ITEMS = len(_fields_)


class NeuronRegions(NamedTuple):
    """
    Identifiers for neuron regions.
    """
    core_params: int
    neuron_params: int
    current_source_params: int
    neuron_recording: int
    neuron_builder: int
    initial_values: int


class PopulationMachineNeurons(
        AbstractNeuronExpandable, allow_derivation=True):
    """
    Mix-in for machine vertices that have neurons in them.
    """

    # This MUST stay empty to allow mixing with other things with slots
    __slots__ = ()

    @property
    @abstractmethod
    def _pop_vertex(self) -> PopulationVertex:
        """
        The application vertex of the machine vertex.

        .. note::
            This is likely to be available via the MachineVertex.
        """
        raise NotImplementedError

    @property
    @abstractmethod
    def _vertex_slice(self) -> Slice:
        """
        The slice of the application vertex atoms on this machine vertex.

        .. note::
            This is likely to be available via the MachineVertex.
        """
        raise NotImplementedError

    @property
    @abstractmethod
    def _slice_index(self) -> int:
        """
        The index of the slice of this vertex in the list of slices.
        """
        raise NotImplementedError

    @property
    @abstractmethod
    def _key(self) -> int:
        """
        The key for spikes.
        """
        raise NotImplementedError

    @property
    @abstractmethod
    def _has_key(self) -> bool:
        """
        Whether a key has been defined.
        """
        raise NotImplementedError

    @abstractmethod
    def _set_key(self, key: int) -> None:
        """
        Set the key for spikes.

        .. note::
            This is required because this class cannot have any storage.

        :param key: The key to be set
        """
        raise NotImplementedError

    @property
    @abstractmethod
    def _neuron_regions(self) -> NeuronRegions:
        """
        The region identifiers for the neuron regions.
        """
        raise NotImplementedError

    @property
    @abstractmethod
    def _neuron_data(self) -> NeuronData:
        """
        The neuron data handler.
        """
        raise NotImplementedError

    @property
    @abstractmethod
    def _max_atoms_per_core(self) -> int:
        """
        The maximum number of atoms on a core, used for neuron data transfer.
        """
        raise NotImplementedError

    @abstractmethod
    def set_do_neuron_regeneration(self) -> None:
        """
        Indicate that data re-generation of neuron parameters is required.
        """
        raise NotImplementedError

    def _parse_neuron_provenance(self, x: int, y: int, p: int,
                                 provenance_data: Sequence[int]) -> None:
        """
        Extract and yield neuron provenance.

        :param x: x coordinate of the chip where this core
        :param y: y coordinate of the core where this core
        :param p: virtual id of the core
        :param provenance_data: A list of data items to interpret
        """
        neuron_prov = NeuronProvenance(*provenance_data)
        with ProvenanceWriter() as db:
            db.insert_core(
                x, y, p, "Last_timer_tic_the_core_ran_to",
                neuron_prov.current_timer_tick)
            db.insert_core(
                x, y, p, "Earliest_send_time", neuron_prov.earliest_send)
            db.insert_core(
                x, y, p, "Latest_Send_time", neuron_prov.latest_send)

    def _write_neuron_data_spec(
            self, spec: DataSpecificationGenerator,
            ring_buffer_shifts: Sequence[int]) -> None:
        """
        Write the data specification of the neuron data.

        :param spec: The data specification to write to
        :param ring_buffer_shifts:
            The shifts to apply to convert ring buffer values to S1615 values
        """
        # Get and store the key
        key = self.__find_default_key()
        if key is not None:
            self._set_key(key)

        # Write the neuron core parameters
        self._write_neuron_core_parameters(spec, ring_buffer_shifts)

        # Write the current source parameters
        self._write_current_source_parameters(spec)

        # Write the other parameters
        self._neuron_data.write_data(
            spec, self._vertex_slice, self._neuron_regions)

    def __find_default_key(self) -> Optional[int]:
        routing_info = SpynnakerDataView.get_routing_infos()
        if not self._pop_vertex.extra_partitions:
            return routing_info.get_single_key_from(
                cast(AbstractVertex, self))
        partition_ids = set(
            routing_info.get_partitions_from(
                cast(AbstractVertex, self)))
        partition_ids = partition_ids - set(self._pop_vertex.extra_partitions)
        if len(partition_ids) > 1:
            raise ValueError(
                "Multiple outgoing partitions found, cannot determine key")
        if len(partition_ids) == 0:
            return None
        return routing_info.get_key_from(
            cast(AbstractVertex, self), next(iter(partition_ids)))

    def _rewrite_neuron_data_spec(
            self, spec: DataSpecificationReloader) -> None:
        """
        Re-Write the data specification of the neuron data.

        :param spec: The data specification to write to
        """
        # Write the current source parameters
        self._write_current_source_parameters(spec)

        # Write the other parameters after forcing a regeneration
        self._neuron_data.write_data(
            spec, self._vertex_slice, self._neuron_regions, False)

    def _write_neuron_core_parameters(
            self, spec: DataSpecificationGenerator,
            ring_buffer_shifts: Sequence[int]) -> None:
        """
        Write the neuron parameters region.

        :param spec: The data specification to write to
        :param ring_buffer_shifts:
            The shifts to apply to convert ring buffer values to S1615 values
        """
        n_atoms = self._vertex_slice.n_atoms
        spec.comment(f"\nWriting Neuron Parameters for {n_atoms} Neurons:\n")

        # Reserve and switch to the memory region
        params_size = self._pop_vertex.get_sdram_usage_for_core_neuron_params(
            n_atoms)
        spec.reserve_memory_region(
            region=self._neuron_regions.core_params, size=params_size,
            label='Neuron Core Params')
        spec.switch_write_focus(self._neuron_regions.core_params)

        # Write whether the key is to be used, and then the key, or 0 if it
        # isn't to be used
        keys: Union[numpy.ndarray, List[int]]
        if not self._has_key:
            spec.write_value(data=0)
            keys = [0] * n_atoms
        else:
            spec.write_value(data=1)
            keys = get_keys(
                self._key, self._vertex_slice, self._pop_vertex.n_colour_bits)

        # Write the number of neurons in the block:
        spec.write_value(data=n_atoms)

        # Write the maximum number of neurons based on the max atoms per core
        spec.write_value(data=2**get_n_bits(self._max_atoms_per_core))

        # Write the number of colour bits
        spec.write_value(self._pop_vertex.n_colour_bits)

        # Write the ring buffer data
        # This is only the synapse types that need a ring buffer i.e. not
        # those stored in synapse dynamics
        n_synapse_types = self._pop_vertex.neuron_impl.get_n_synapse_types()
        spec.write_value(n_synapse_types)
        spec.write_array(ring_buffer_shifts)

        # Write the keys
        spec.write_array(keys)

    def __in_selector(
            self, n: Union[int, numpy.integer], selector: Selector) -> bool:
        if isinstance(selector, Container):
            return n in selector
        return n == selector

    def _write_current_source_parameters(
            self, spec: DataSpecificationBase) -> None:
        n_atoms: int = self._vertex_slice.n_atoms

        spec.comment(
            f"\nWriting Current Source Parameters for {n_atoms} Neurons:\n")

        # Reserve and switch to the current source region
        params_size = self._pop_vertex.\
            get_sdram_usage_for_current_source_params(n_atoms)
        spec.reserve_memory_region(
            region=self._neuron_regions.current_source_params,
            size=params_size, label='CurrentSourceParams')
        spec.switch_write_focus(self._neuron_regions.current_source_params)

        # Get the current sources from the app vertex
        current_source_id_list = self._pop_vertex.current_source_id_list

        # Work out which current sources are on this core
        current_sources = self.__get_current_sources_sorted()

        # Write the number of sources
        spec.write_value(len(current_sources))

        # Don't write anything else if there are no current sources
        if current_sources:
            # Array to keep track of the number of each type of current source
            # (there are four, but they are numbered 1 to 4, so five elements)
            cs_index_array: List[int] = [0, 0, 0, 0, 0]

            # Data sent to the machine will be current sources per neuron
            # This array will have the first entry indicating the number of
            # sources for each neuron, then if this is non-zero, follow it with
            # the IDs indicating the current source ID value, and then the
            # index within that type of current source
            neuron_current_sources = [[0] for n in range(n_atoms)]
            for current_source in current_sources:
                # Get the ID of the current source
                cs_id = current_source.current_source_id

                # Only use IDs that are on this core
                for i, raster_id in enumerate(
                        self._vertex_slice.get_raster_ids()):
                    if self.__in_selector(
                            raster_id, current_source_id_list[current_source]):
                        # I think this is now right, but test it more...
                        neuron_current_sources[i][0] += 1
                        neuron_current_sources[i].append(cs_id)
                        neuron_current_sources[i].append(
                            cs_index_array[cs_id])

                # Increase the ID value in case a (different) current source
                # of the same type is also used
                cs_index_array[cs_id] += 1

            # Now loop over the neurons on this core and write the current
            # source ID and index for sources attached to each neuron
            for atom in range(n_atoms):
                n_current_sources = neuron_current_sources[atom][0]
                spec.write_value(n_current_sources)
                if n_current_sources != 0:
                    for csid in range(n_current_sources * 2):
                        spec.write_value(neuron_current_sources[atom][csid+1])

            # Write the number of each type of current source
            for cs_index in range(1, len(cs_index_array)):
                spec.write_value(cs_index_array[cs_index])

            # Now loop over the current sources and write the data required
            # for each type of current source
            for current_source in current_sources:
                cs_data_types = current_source.parameter_types
                cs_id = current_source.current_source_id
                for key, value in current_source.parameters.items():
                    # StepCurrentSource currently handled with arrays
                    if cs_id == CurrentSourceIDs.STEP_CURRENT_SOURCE.value:
                        assert isinstance(value, Sequence)
                        n_params = len(value)
                        spec.write_value(n_params)
                        for n_p in range(n_params):
                            value_convert = convert_to(
                                value[n_p], cs_data_types[key]).view("uint32")
                            spec.write_value(data=value_convert)
                    # All other sources have single-valued params
                    else:
                        if isinstance(value, Sequence):
                            for a_value in value:
                                value_convert = convert_to(
                                    a_value,
                                    cs_data_types[key]).item()
                                spec.write_value(data=value_convert)
                        else:
                            value_convert = convert_to(
                                value, cs_data_types[key]).item()
                            spec.write_value(data=value_convert)

    def __get_current_sources_sorted(self) -> List[AbstractCurrentSource]:
        app_current_sources = self._pop_vertex.current_sources
        current_source_id_list = self._pop_vertex.current_source_id_list

        current_sources: Set[AbstractCurrentSource] = set()
        for app_current_source in app_current_sources:
            for n in self._vertex_slice.get_raster_ids():
                if self.__in_selector(
                        n, current_source_id_list[app_current_source]):
                    current_sources.add(app_current_source)

        # Sort the current sources into current_source_id order
        return sorted(
            current_sources, key=lambda x: x.current_source_id)

    def read_parameters_from_machine(self, placement: Placement) -> None:
        """
        Read the parameters and state of the neurons from the machine
        at the current time.

        :param placement: Where to read the data from
        """
        self._neuron_data.read_data(placement, self._neuron_regions)

    def read_initial_parameters_from_machine(
            self, placement: Placement) -> None:
        """
        Read the parameters and state of the neurons from the machine
        as they were at the last time 0.

        :param placement: Where to read the data from
        """
        self._neuron_data.read_initial_data(placement, self._neuron_regions)

    @overrides(AbstractNeuronExpandable.gen_neurons_on_machine)
    def gen_neurons_on_machine(self) -> bool:
        return self._neuron_data.gen_on_machine

    @property
    @overrides(AbstractNeuronExpandable.neuron_generator_region)
    def neuron_generator_region(self) -> int:
        return self._neuron_regions.neuron_builder

    @overrides(AbstractNeuronExpandable.read_generated_initial_values)
    def read_generated_initial_values(self, placement: Placement) -> None:
        # Only do this if we actually need the data now i.e. if someone has
        # requested that the data be read before calling run
        if self._pop_vertex.read_initial_values:
            # If we do decide to read now, we can also copy the initial values
            self._neuron_data.read_data(placement, self._neuron_regions)
            self._pop_vertex.copy_initial_state_variables(self._vertex_slice)



# Copyright (c) 2014 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from .connection_holder_finisher import finish_connection_holders
from .redundant_packet_count_report import redundant_packet_count_report
from .spynnaker_connection_holder_generations import (
    SpYNNakerConnectionHolderGenerator)
from .spynnaker_neuron_network_specification_report import (
    spynnaker_neuron_graph_network_specification_report)
from .spynnaker_synaptic_matrix_report import SpYNNakerSynapticMatrixReport
from .synapse_expander import synapse_expander
from .delay_support_adder import delay_support_adder
from .neuron_expander import neuron_expander

__all__ = [
    "delay_support_adder",
    "finish_connection_holders",
    "redundant_packet_count_report",
    "SpYNNakerConnectionHolderGenerator",
    "spynnaker_neuron_graph_network_specification_report",
    "SpYNNakerSynapticMatrixReport",
    "synapse_expander", "neuron_expander"]



# Copyright (c) 2015 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from spinn_utilities.overrides import overrides
from spinn_utilities.ranged.range_dictionary import RangeDictionary

from spinn_front_end_common.interface.ds import DataType

from spynnaker.pyNN.models.neuron.implementations import ModelParameter
from spynnaker.pyNN.utilities.struct import Struct

from .abstract_input_type import AbstractInputType

E_REV_E = "e_rev_E"
E_REV_I = "e_rev_I"


class InputTypeConductance(AbstractInputType):
    """
    The conductance input type.
    """
    __slots__ = (
        "__e_rev_E",
        "__e_rev_I")

    def __init__(self, e_rev_E: ModelParameter, e_rev_I: ModelParameter):
        """
        :param e_rev_E: Reversal potential for excitatory input;
            :math:`E^{rev}_e`
        :param e_rev_I: Reversal potential for inhibitory input;
            :math:`E^{rev}_i`
        """
        super().__init__(
            [Struct([(DataType.S1615, E_REV_E),
                     (DataType.S1615, E_REV_I)])],
            {E_REV_E: "mV", E_REV_I: "mV"})
        # pylint: disable=invalid-name
        self.__e_rev_E = e_rev_E
        self.__e_rev_I = e_rev_I

    @overrides(AbstractInputType.add_parameters)
    def add_parameters(self, parameters: RangeDictionary[float]) -> None:
        parameters[E_REV_E] = self._convert(self.__e_rev_E)
        parameters[E_REV_I] = self._convert(self.__e_rev_I)

    @overrides(AbstractInputType.add_state_variables)
    def add_state_variables(
            self, state_variables: RangeDictionary[float]) -> None:
        pass

    @overrides(AbstractInputType.get_global_weight_scale)
    def get_global_weight_scale(self) -> float:
        return 1024.0

    @property
    def e_rev_E(self) -> ModelParameter:
        # pylint: disable=invalid-name
        """
        :math:`E_{{rev}_e}`
        """
        return self.__e_rev_E

    @property
    def e_rev_I(self) -> ModelParameter:
        # pylint: disable=invalid-name
        """
        :math:`E_{{rev}_i}`
        """
        return self.__e_rev_I



# Copyright (c) 2014 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations
from dataclasses import dataclass
import logging
from typing import (
    Dict, List, Optional, Sequence, Tuple, Union, cast, TYPE_CHECKING)

import numpy
from numpy import floating, integer, int64, uint32
from numpy.typing import NDArray
from typing_extensions import TypeGuard

from spinn_utilities.overrides import overrides
from spinn_utilities.log import FormatAdapter

from pacman.model.graphs import AbstractVertex
from pacman.model.graphs.application import ApplicationVertex
from pacman.model.graphs.machine import MachineVertex
from pacman.model.graphs.common import Slice

from spynnaker.pyNN.data import SpynnakerDataView
from spynnaker.pyNN.exceptions import InvalidParameterType
from spynnaker.pyNN.types import DELAYS, WEIGHTS_DELAYS, WEIGHTS
from .abstract_connector import AbstractConnector
from .abstract_generate_connector_on_host import (
    AbstractGenerateConnectorOnHost)

if TYPE_CHECKING:
    from spynnaker.pyNN.models.neural_projections import (
        ProjectionApplicationEdge, SynapseInformation)
    from spynnaker.pyNN.models.neuron.synapse_dynamics import (
        AbstractSynapseDynamics)

logger = FormatAdapter(logging.getLogger(__name__))

# Indices of the source and target in the connection list array
_SOURCE = 0
_TARGET = 1
_FIRST_PARAM = 2


def _is_sequential(value: WEIGHTS_DELAYS
                   ) -> TypeGuard[NDArray[numpy.float64]]:
    return isinstance(value, numpy.ndarray)


@dataclass(frozen=True)
class _ExtraParams:
    """
    Information about extra parameters.
    """
    data: NDArray
    names: Sequence[str]


class FromListConnector(AbstractConnector, AbstractGenerateConnectorOnHost):
    """
    Make connections according to a list.
    """
    __slots__ = (
        "__conn_list",
        "__column_names",
        "__sources",
        "__targets",
        "__weights",
        "__delays",
        "__extra_params",
        "__split_conn_list",
        "__split_post_slices")

    def __init__(self, conn_list: Union[None, NDArray, List[Tuple[int, ...]]],
                 column_names: Optional[Sequence[str]] = None, *,
                 safe: bool = True, verbose: bool = False,
                 callback: None = None):
        """
        :param conn_list:
            A numpy array or a list of tuples, one tuple for each connection.
            Each tuple should contain::

                (pre_idx, post_idx, p1, p2, ..., pn)

            where ``pre_idx`` is the index (i.e. order in the Population,
            not the ID) of the presynaptic neuron, ``post_idx`` is
            the index of the postsynaptic neuron, and
            ``p1``, ``p2``, etc. are the synaptic parameters (e.g.,
            weight, delay, plasticity parameters).
            All tuples/rows must have the same number of items.
        :param safe:
            if ``True``, check that weights and delays have valid values.
            If ``False``, this check is skipped.
        :param verbose:
            Whether to output extra information about the connectivity to a
            CSV file
        :param column_names: the names of the parameters ``p1``, ``p2``, etc.
            If not provided, it is assumed the parameters are ``weight, delay``
            (for backwards compatibility).
        :param callback:
            if given, a callable that display a progress bar on the terminal.

            .. note::
                Not supported by sPyNNaker.
        """
        super().__init__(safe, callback, verbose)

        self.__column_names = column_names
        self.__split_conn_list: Dict[int, NDArray[integer]] = {}
        self.__split_post_slices: Optional[List[Slice]] = None

        # These are set by the conn_list setter
        self.__conn_list: NDArray
        self.__sources: NDArray[uint32]
        self.__targets: NDArray[uint32]
        self.__delays: Optional[NDArray[floating]]
        self.__weights: Optional[NDArray[floating]]
        self.__extra_params: Optional[_ExtraParams]

        # Call the conn_list setter, as this sets the internal values
        self.conn_list = cast(NDArray, conn_list)  # Bug: python/mypy#3004

    @overrides(AbstractConnector.get_delay_maximum)
    def get_delay_maximum(self, synapse_info: SynapseInformation) -> float:
        if self.__delays is None:
            if _is_sequential(synapse_info.delays):
                return float(numpy.max(synapse_info.delays))
            return self._get_delay_maximum(
                synapse_info.delays, len(self.__targets), synapse_info)
        else:
            return numpy.max(self.__delays)

    @overrides(AbstractConnector.get_delay_minimum)
    def get_delay_minimum(self, synapse_info: SynapseInformation) -> float:
        if self.__delays is None:
            if _is_sequential(synapse_info.delays):
                return float(numpy.min(synapse_info.delays))
            return self._get_delay_minimum(
                synapse_info.delays, len(self.__targets), synapse_info)
        else:
            return numpy.min(self.__delays)

    @overrides(AbstractConnector.get_delay_variance)
    def get_delay_variance(self, delays: DELAYS,
                           synapse_info: SynapseInformation) -> float:
        if self.__delays is None:
            if _is_sequential(synapse_info.delays):
                return float(numpy.var(synapse_info.delays))
            return AbstractConnector.get_delay_variance(
                self, delays, synapse_info)
        else:
            return numpy.var(self.__delays)

    def __id_to_m_vertex_index(
            self, n_atoms: int, slices: Sequence[Slice]) -> NDArray[integer]:
        """ Produce an array that maps from a vertex atom id to the
            machine vertex index of the target
        """
        mapping = numpy.zeros(n_atoms, dtype="uint32")
        for i, s in enumerate(slices):
            mapping[s.get_raster_ids()] = i
        return mapping

    def _split_connections(
            self, n_pre_atoms: int, n_post_atoms: int,
            post_slices: Sequence[Slice]) -> \
            Tuple[NDArray[integer], NDArray[integer],
                  Optional[NDArray[floating]],
                  Optional[NDArray[floating]]]:
        input_filter = numpy.logical_and(
            self.__targets < n_post_atoms, self.__sources < n_pre_atoms)
        targets = self.__targets[input_filter]
        sources = self.__sources[input_filter]
        weights = (None if self.__weights is None else
                   self.__weights[input_filter])
        delays = (None if self.__delays is None else
                  self.__delays[input_filter])

        # If nothing has changed, use the cache
        if self.__split_post_slices == post_slices:
            return sources, targets, weights, delays

        # If there are no connections, return
        if len(self.__conn_list) == 0:
            self.__split_conn_list = {}
            return sources, targets, weights, delays

        self.__split_post_slices = list(post_slices)
        m_vertex_mapping = self.__id_to_m_vertex_index(
            n_post_atoms, post_slices)

        # Get which index of the from list is on which vertex
        target_vertices = m_vertex_mapping[self.__targets[input_filter]]

        # Get how many on each vertex there are
        index_count = numpy.bincount(
            target_vertices, minlength=len(post_slices))[:len(post_slices)]

        # Get the sort order to put connections on each vertex
        sort_indices = numpy.argsort(target_vertices)

        # Split the sort order in to groups of connection indices
        split_indices = numpy.array(numpy.split(
            sort_indices, numpy.cumsum(index_count)), dtype=object)[:-1]

        # Get the results indexed by hi_atom in the slices
        self.__split_conn_list = {
            post_slices[i].lo_atom: indices
            for i, indices in enumerate(split_indices)
            if len(indices) > 0
        }
        return sources, targets, weights, delays

    @overrides(AbstractConnector.get_n_connections_from_pre_vertex_maximum)
    def get_n_connections_from_pre_vertex_maximum(
            self, n_post_atoms: int, synapse_info: SynapseInformation,
            min_delay: Optional[float] = None,
            max_delay: Optional[float] = None) -> int:
        mask = None
        delays_handled = False
        if (min_delay is not None and max_delay is not None):
            if self.__delays is not None:
                mask = ((self.__delays >= min_delay) &
                        (self.__delays <= max_delay))
                delays_handled = True
            elif _is_sequential(synapse_info.delays):
                delays = synapse_info.delays
                mask = ((delays >= min_delay) & (delays <= max_delay))
                delays_handled = True
        if mask is None:
            conns = self.__conn_list.copy()
        else:
            conns = self.__conn_list[mask].copy()
        if conns.size == 0:
            return 0

        # Make targets be core indices
        conns[:, _TARGET] //= n_post_atoms

        # Split into sources
        source_split_conns = self.__numpy_group(conns, _SOURCE)

        # Split into groups by post_n_atoms
        target_split_conns = [
            self.__numpy_group(s, _TARGET) for s in source_split_conns]

        # Find the biggest group
        max_targets = max([len(t) for s in target_split_conns for t in s])

        # If no delays just return max targets as this is for all delays
        # If there are delays in the list, this was also handled above
        if min_delay is None or max_delay is None or delays_handled:
            return max_targets

        # If here, there must be no delays in the list, so use the passed in
        # ones
        return self._get_n_connections_from_pre_vertex_with_delay_maximum(
            synapse_info.delays,
            synapse_info.n_pre_neurons * synapse_info.n_post_neurons,
            max_targets, min_delay, max_delay, synapse_info)

    @staticmethod
    def __numpy_group(conns: NDArray, column: int) -> List[NDArray]:
        # Sort by the column to group by
        s = conns[conns[:, column].argsort()]

        # Find split points by getting the first indices of the unique items
        # and then removing the first (as that will be 0 and we don't want to
        # split at 0)
        split_points = numpy.unique(s[:, column], return_index=True)[1][1:]

        # Perform the split
        return numpy.array_split(conns, split_points)

    @overrides(AbstractConnector.get_n_connections_to_post_vertex_maximum)
    def get_n_connections_to_post_vertex_maximum(
            self, synapse_info: SynapseInformation) -> int:
        if len(self.__targets) == 0:
            return 0
        return int(numpy.max(numpy.bincount(
            self.__targets.astype(int64, copy=False))))

    @overrides(AbstractConnector.get_weight_mean)
    def get_weight_mean(self, weights: WEIGHTS,
                        synapse_info: SynapseInformation) -> float:
        if self.__weights is None:
            if _is_sequential(synapse_info.weights):
                return float(numpy.mean(synapse_info.weights))
            return AbstractConnector.get_weight_mean(
                self, weights, synapse_info)
        else:
            return float(numpy.mean(numpy.abs(self.__weights)))

    @overrides(AbstractConnector.get_weight_maximum)
    def get_weight_maximum(self, synapse_info: SynapseInformation) -> float:
        if self.__weights is None:
            if _is_sequential(synapse_info.weights):
                return float(numpy.amax(synapse_info.weights))
            return self._get_weight_maximum(
                synapse_info.weights, len(self.__targets), synapse_info)
        else:
            return float(numpy.amax(numpy.abs(self.__weights)))

    @overrides(AbstractConnector.get_weight_variance)
    def get_weight_variance(self, weights: WEIGHTS,
                            synapse_info: SynapseInformation) -> float:
        if self.__weights is None:
            if _is_sequential(synapse_info.weights):
                return float(numpy.var(synapse_info.weights))
            return AbstractConnector.get_weight_variance(
                self, weights, synapse_info)
        else:
            return float(numpy.var(numpy.abs(self.__weights)))

    @overrides(AbstractGenerateConnectorOnHost.create_synaptic_block)
    def create_synaptic_block(
            self, post_slices: Sequence[Slice], post_vertex_slice: Slice,
            synapse_type: int, synapse_info: SynapseInformation) -> NDArray:
        sources, targets, weights, delays = self._split_connections(
            synapse_info.n_pre_neurons, synapse_info.n_post_neurons,
            post_slices)
        post_lo = post_vertex_slice.lo_atom
        if post_lo not in self.__split_conn_list:
            return numpy.zeros(0, dtype=self.NUMPY_SYNAPSES_DTYPE)
        else:
            indices = self.__split_conn_list[post_lo]

        block = numpy.zeros(len(indices), dtype=self.NUMPY_SYNAPSES_DTYPE)
        block["source"] = synapse_info.pre_vertex.get_key_ordered_indices(
            sources[indices])
        block["target"] = post_vertex_slice.get_relative_indices(
            targets[indices])
        # check that conn_list has weights, if not then use the value passed in
        if weights is None:
            if _is_sequential(synapse_info.weights):
                block["weight"] = numpy.array(synapse_info.weights)[indices]
            else:
                block["weight"] = self._generate_weights(
                    block["source"], block["target"], len(indices),
                    post_vertex_slice, synapse_info)
        else:
            block["weight"] = weights[indices]
        # check that conn_list has delays, if not then use the value passed in
        if delays is None:
            if _is_sequential(synapse_info.delays):
                block["delay"] = numpy.array(synapse_info.delays)[indices]
            else:
                block["delay"] = self._generate_delays(
                    block["source"], block["target"], len(indices),
                    post_vertex_slice, synapse_info)
        else:
            block["delay"] = self._clip_delays(delays[indices])
        block["synapse_type"] = synapse_type
        return block

    def __repr__(self) -> str:
        return f"FromListConnector(n_connections={len(self.__sources)})"

    @property
    def conn_list(self) -> NDArray:
        """
        The connection list.
        """
        return self.__conn_list

    @conn_list.setter
    def conn_list(self, conn_list: Union[
            None, NDArray, List[Tuple[int, ...]]]) -> None:
        if conn_list is None or len(conn_list) == 0:
            self.__conn_list = numpy.zeros((0, 2), dtype=uint32)
        else:
            self.__conn_list = numpy.array(conn_list)

        # If the shape of the conn_list is 2D, numpy has been able to create
        # a 2D array which means every entry has the same number of values.
        # If this was not possible, raise an exception!
        if len(self.__conn_list.shape) != 2:
            raise InvalidParameterType(
                "Each tuple in the connection list for the"
                " FromListConnector must have the same number of elements")

        # This tells us how many columns are in the list
        n_columns = self.__conn_list.shape[1]
        if n_columns < 2:
            raise InvalidParameterType(
                "Each tuple in the connection list for the"
                " FromListConnector must have at least 2 elements")
        if (self.__column_names is not None and
                n_columns != len(self.__column_names) + _FIRST_PARAM):
            raise InvalidParameterType(
                "The number of column names must match the number of"
                " additional elements in each tuple in the connection list,"
                " not including the pre_idx or post_idx")

        # Get the column names if not specified
        if self.__column_names is not None:
            column_names = self.__column_names
        else:
            if n_columns == 4:
                column_names = ('weight', 'delay')
            elif n_columns == 2:
                column_names = ()
            else:
                raise TypeError(
                    f"Need to set 'column_names' for n_columns={n_columns}")

        # Set the source and targets
        self.__sources = self.__conn_list[:, _SOURCE].astype(numpy.uint32)
        self.__targets = self.__conn_list[:, _TARGET].astype(numpy.uint32)

        # Find any weights
        try:
            weight_column = column_names.index('weight') + _FIRST_PARAM
            self.__weights = self.__conn_list[:, weight_column]
        except ValueError:
            self.__weights = None

        # Find any delays
        try:
            delay_column = column_names.index('delay') + _FIRST_PARAM
            self.__delays = (numpy.rint(
                numpy.array(self.__conn_list[:, delay_column]) *
                SpynnakerDataView.get_simulation_time_step_per_ms()) *
                SpynnakerDataView.get_simulation_time_step_ms())
        except ValueError:
            self.__delays = None

        # Find extra columns
        extra_columns = [
            i + _FIRST_PARAM
            for i, name in enumerate(column_names)
            if name not in ('weight', 'delay')]

        # Check any additional parameters have single values over the whole
        # set of connections (as other things aren't currently supported)
        for i in extra_columns:
            # pylint: disable=wrong-spelling-in-comment
            # numpy.ptp gives the difference between the maximum and
            # minimum values of an array, so if 0, all values are equal
            if numpy.ptp(self.__conn_list[:, i]):
                raise ValueError(
                    f"All values in column {i} "
                    f"({column_names[i - _FIRST_PARAM]}) of a "
                    "FromListConnector must have the same value")

        # Store the extra data
        if extra_columns:
            self.__extra_params = _ExtraParams(
                self.__conn_list[:, extra_columns],
                tuple(
                    column_names[i - _FIRST_PARAM] for i in extra_columns))
        else:
            self.__extra_params = None

    @property
    def column_names(self) -> Optional[Sequence[str]]:
        """
        The names of the columns in the array after the first two.
        Of particular interest is whether ``weight`` and ``delay`` columns
        are present.
        """
        return self.__column_names

    @column_names.setter
    def column_names(self, column_names: Optional[Sequence[str]]) -> None:
        self.__column_names = column_names

    def get_extra_parameters(self) -> Optional[NDArray]:
        """
        Getter for the extra parameters. Excludes ``weight`` and
        ``delay`` columns.

        :return: The extra parameters
        """
        return self.__extra_params.data if self.__extra_params else None

    def get_extra_parameter_names(self) -> Optional[Sequence[str]]:
        """
        :returns: The names of the extra parameters or None if there are None
        """
        return self.__extra_params.names if self.__extra_params else None

    @overrides(AbstractConnector.get_connected_vertices)
    def get_connected_vertices(
            self, s_info: SynapseInformation,
            source_vertex: ApplicationVertex,
            target_vertex: ApplicationVertex) -> Sequence[
                Tuple[MachineVertex, Sequence[AbstractVertex]]]:
        # Divide the targets into bins based on post slices
        post_slices = [m.vertex_slice
                       for m in target_vertex.splitter.get_in_coming_vertices(
                           s_info.partition_id)]
        pre_vertices = source_vertex.splitter.get_out_going_vertices(
            s_info.partition_id)
        pre_slices = [m.vertex_slice for m in pre_vertices]

        post_mapping = self.__id_to_m_vertex_index(
            s_info.n_post_neurons, post_slices)
        pre_mapping = self.__id_to_m_vertex_index(
            s_info.n_pre_neurons, pre_slices)

        input_filter = numpy.logical_and(
            self.__targets < target_vertex.n_atoms,
            self.__sources < source_vertex.n_atoms)

        target_vertices = post_mapping[self.__targets[input_filter]]
        source_vertices = pre_mapping[self.__sources[input_filter]]

        # Join the groups from both axes
        n_bins = (len(pre_slices), len(post_slices))
        joined_indices = numpy.ravel_multi_index(
            (source_vertices, target_vertices), n_bins, mode="clip")

        # Get a count of the indices in each bin
        index_count = numpy.bincount(
            joined_indices, minlength=numpy.prod(n_bins))

        pre_post_lo = [(pre.lo_atom, post.lo_atom) for pre in pre_slices
                       for post in post_slices]

        # Put the counts into a dict by hi-atom
        split_counts = {
            pre_post: count
            for pre_post, count in zip(pre_post_lo, index_count)
            if count > 0
        }

        return [
            (m_vert, [s_vert for s_vert in pre_vertices
                      if (s_vert.vertex_slice.lo_atom,
                          m_vert.vertex_slice.lo_atom) in split_counts])
            for m_vert in target_vertex.splitter.get_in_coming_vertices(
                s_info.partition_id)
        ]

    def _apply_parameters_to_synapse_type(
            self, synapse_type: AbstractSynapseDynamics) -> None:
        if self.__extra_params:
            for i, name in enumerate(self.__extra_params.names):
                synapse_type.set_value(name, self.__extra_params.data[:, i])

    @overrides(AbstractConnector.validate_connection)
    def validate_connection(
            self, application_edge: ProjectionApplicationEdge,
            synapse_info: SynapseInformation) -> None:
        out_of_range_targets = self.__targets >= synapse_info.n_post_neurons
        if any(out_of_range_targets):
            logger.warning(
                "Some targets are out of range: {}",
                self.__conn_list[out_of_range_targets])
        out_of_range_sources = self.__sources >= synapse_info.n_pre_neurons
        if any(out_of_range_sources):
            logger.warning(
                "Some sources are out of range: {}",
                self.__conn_list[out_of_range_sources])



# Copyright (c) 2014 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from spinn_front_end_common.utilities.constants import (
    BYTES_PER_WORD, BYTES_PER_KB)

POSSION_SIGMA_SUMMATION_LIMIT = 3.0

BLOCK_INDEX_HEADER_WORDS = 3
BLOCK_INDEX_ROW_WORDS = 2

RECORD_SPIKE_BIT = 1 << 0
RECORD_STATE_BIT = 1 << 1
RECORD_GSYN_BIT = 1 << 2
RECORDING_ENTRY_BYTE_SIZE = BYTES_PER_WORD

# pylint: disable=wrong-spelling-in-comment
# From neuron common-typedefs.h
SYNAPSE_INDEX_BITS = 8
MAX_NEURON_SIZE = 1 << SYNAPSE_INDEX_BITS
#: The size of each output spike line
OUT_SPIKE_SIZE = MAX_NEURON_SIZE >> 5
#: The number of bytes for each spike line
OUT_SPIKE_BYTES = OUT_SPIKE_SIZE * BYTES_PER_WORD
V_BUFFER_SIZE_PER_TICK_PER_NEURON = BYTES_PER_WORD
GSYN_BUFFER_SIZE_PER_TICK_PER_NEURON = 2 * BYTES_PER_WORD

SPIKE_BUFFER_SIZE_BUFFERING_IN = 1 * 1024 * BYTES_PER_KB
EIEIO_SPIKE_BUFFER_SIZE_BUFFERING_OUT = 1 * 1024 * BYTES_PER_KB
EIEIO_BUFFER_SIZE_BEFORE_RECEIVE = 512 * BYTES_PER_KB

INFINITE_SIMULATION = 4294967295

# from synaptic manager
#: Words: 2 for row length and number of rows and 1 for plastic region size
#: (which might be 0)
SYNAPTIC_ROW_HEADER_WORDS = 2 + 1

NA_TO_PA_SCALE = 1000.0

# might not be used
WEIGHT_FLOAT_TO_FIXED_SCALE = 16.0
SCALE = WEIGHT_FLOAT_TO_FIXED_SCALE * NA_TO_PA_SCALE

#: the minimum supported delay slot between two neurons
MIN_SUPPORTED_DELAY = 1

#: The partition ID used for spike data
SPIKE_PARTITION_ID = "SPIKE"

# names for recording components
SPIKES = 'spikes'
MEMBRANE_POTENTIAL = "v"
GSYN_EXCIT = "gsyn_exc"
GSYN_INHIB = "gsyn_inh"
REWIRING = "rewiring"

#: The partition ID used for Poisson live control data
LIVE_POISSON_CONTROL_PARTITION_ID = "CONTROL"

#: The maximum row length of the master population table
POP_TABLE_MAX_ROW_LENGTH = 256

#: The name of the partition for Synaptic SDRAM
SYNAPSE_SDRAM_PARTITION_ID = "SDRAM Synaptic Inputs"

#: The conservative amount of write bandwidth available on a chip
WRITE_BANDWIDTH_BYTES_PER_SECOND = 250 * 1024 * 1024

#: The maximum number of bits for the ring buffer index that are likely to
#: fit in DTCM (14-bits = 16,384 16-bit ring buffer entries = 32Kb DTCM
MAX_RING_BUFFER_BITS = 14



# Copyright (c) 2020 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from spinn_utilities.abstract_base import AbstractBase, abstractmethod


class AbstractSpynnakerSplitterDelay(object, metaclass=AbstractBase):
    """
    Defines that a splitter is able to handle delays in some way.

    Ideally the splitter, and therefore the vertices it creates, are able to
    handle some delay themselves and if more is needed have the ability to
    accept spikes from a :py:class:`DelayExtensionMachineVertex`.
    """

    __slots__ = ()

    @abstractmethod
    def max_support_delay(self) -> int:
        """
        returns the max amount of delay this post vertex can support.

        :return: max delay supported in ticks
        """
        raise NotImplementedError

    def accepts_edges_from_delay_vertex(self) -> bool:
        """
        Confirms that the splitter's vertices can handle spikes coming from a
        :py:class:`DelayExtensionMachineVertex`.

        If this method returns False and the users ask for a delay larger than
        that allowed by :py:meth:`max_support_delay`, an exception will be
        raised saying a different splitter is required.

        :returns: True if extra delays can be supported using delay vertices
        """
        return True



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Optional, Sequence

from spinn_utilities.overrides import overrides

from spinnman.model.enums import ExecutableType

from pacman.model.graphs.application import ApplicationVertex
from pacman.model.graphs.common import Slice
from pacman.model.graphs.machine import MachineVertex
from pacman.model.placements import Placement
from pacman.model.resources import ConstantSDRAM

from spinn_front_end_common.abstract_models import (
    AbstractHasAssociatedBinary)
from spinn_front_end_common.abstract_models import (
    AbstractGeneratesDataSpecification)
from spinn_front_end_common.interface.ds import DataSpecificationGenerator
from spinn_front_end_common.interface.provenance import (
    ProvidesProvenanceDataFromMachineImpl, ProvenanceWriter)
from spinn_front_end_common.interface.simulation import simulation_utilities
from spinn_front_end_common.utilities.constants import (
    SYSTEM_BYTES_REQUIREMENT, SIMULATION_N_BYTES, BYTES_PER_WORD)

from spynnaker.pyNN.data import SpynnakerDataView


class MachineMunichMotorDevice(
        MachineVertex, AbstractGeneratesDataSpecification,
        AbstractHasAssociatedBinary,
        ProvidesProvenanceDataFromMachineImpl):
    """
    An Omnibot motor control device. This has a real vertex and an
    external device vertex.
    """
    __slots__ = (
        "__continue_if_not_different",
        "__delay_time",
        "__delta_threshold",
        "__sample_time",
        "__speed",
        "__update_time")

    MOTOR_PARTITION_ID = "MOTOR"

    # By asking for this number of keys, we will get a mask of 0xFFFFF800,
    # which works with the motor control protocol correctly
    _MOTOR_N_KEYS = 2048

    _N_ATOMS = 6

    _SYSTEM_REGION = 0
    _PARAMS_REGION = 1
    _PROVENANCE_REGION = 2

    _PROVENANCE_ELEMENTS = 1

    _PARAMS_SIZE = 7 * BYTES_PER_WORD

    #: The name of the provenance item saying that packets were lost.
    INPUT_BUFFER_FULL_NAME = "Times_the_input_buffer_lost_packets"

    def __init__(
            self, speed: int, sample_time: int, update_time: int,
            delay_time: int, delta_threshold: int,
            continue_if_not_different: bool, label: Optional[str] = None,
            app_vertex: Optional[ApplicationVertex] = None):
        """
        :param speed:
        :param sample_time:
        :param update_time:
        :param delay_time:
        :param delta_threshold:
        :param continue_if_not_different:
        :param label:
        :param app_vertex:
        """
        super().__init__(
            label=label, app_vertex=app_vertex,
            vertex_slice=Slice(0, self._N_ATOMS - 1))
        self.__speed = speed
        self.__sample_time = sample_time
        self.__update_time = update_time
        self.__delay_time = delay_time
        self.__delta_threshold = delta_threshold
        self.__continue_if_not_different = bool(continue_if_not_different)

    @property
    @overrides(MachineVertex.sdram_required)
    def sdram_required(self) -> ConstantSDRAM:
        return ConstantSDRAM(
                SYSTEM_BYTES_REQUIREMENT + self._PARAMS_SIZE +
                self.get_provenance_data_size(self._PROVENANCE_ELEMENTS))

    @overrides(AbstractHasAssociatedBinary.get_binary_file_name)
    def get_binary_file_name(self) -> str:
        return "robot_motor_control.aplx"

    @overrides(AbstractHasAssociatedBinary.get_binary_start_type)
    def get_binary_start_type(self) -> ExecutableType:
        return ExecutableType.USES_SIMULATION_INTERFACE

    @property
    @overrides(ProvidesProvenanceDataFromMachineImpl._provenance_region_id)
    def _provenance_region_id(self) -> int:
        return self._PROVENANCE_REGION

    @property
    @overrides(ProvidesProvenanceDataFromMachineImpl._n_additional_data_items)
    def _n_additional_data_items(self) -> int:
        return self._PROVENANCE_ELEMENTS

    @overrides(
        ProvidesProvenanceDataFromMachineImpl.parse_extra_provenance_items)
    def parse_extra_provenance_items(
            self, label: str, x: int, y: int, p: int,
            provenance_data: Sequence[int]) -> None:
        n_buffer_overflows, = provenance_data

        with ProvenanceWriter() as db:
            db.insert_core(x, y, p, self.INPUT_BUFFER_FULL_NAME,
                           n_buffer_overflows)
            if n_buffer_overflows > 0:
                db.insert_report(
                    f"The input buffer for {label} lost packets on "
                    f"{n_buffer_overflows} occasions. "
                    "This is often a sign that the system is running too "
                    "quickly for the number of neurons per core.  "
                    "Please increase the timer_tic or time_scale_factor "
                    "or decrease the number of neurons per core.")

    @overrides(AbstractGeneratesDataSpecification.generate_data_specification)
    def generate_data_specification(self, spec: DataSpecificationGenerator,
                                    placement: Placement) -> None:
        # reserve regions
        self.reserve_memory_regions(spec)

        # Write the setup region
        spec.comment("\n*** Spec for robot motor control ***\n\n")

        # handle simulation data
        spec.switch_write_focus(self._SYSTEM_REGION)
        vertex = placement.vertex
        assert isinstance(vertex, AbstractHasAssociatedBinary)
        spec.write_array(simulation_utilities.get_simulation_header_array(
            vertex.get_binary_file_name()))

        # Get the key
        routing_info = SpynnakerDataView.get_routing_infos()
        edge_key = routing_info.get_key_from(
            placement.vertex, self.MOTOR_PARTITION_ID)

        # write params to memory
        spec.switch_write_focus(region=self._PARAMS_REGION)
        spec.write_value(data=edge_key)
        spec.write_value(data=self.__speed)
        spec.write_value(data=self.__sample_time)
        spec.write_value(data=self.__update_time)
        spec.write_value(data=self.__delay_time)
        spec.write_value(data=self.__delta_threshold)
        spec.write_value(data=int(self.__continue_if_not_different))

        # End-of-Spec:
        spec.end_specification()

    def reserve_memory_regions(self, spec: DataSpecificationGenerator) -> None:
        """
        Reserve SDRAM space for memory areas:

        #. Area for information on what data to record
        #. area for start commands
        #. area for end commands

        :param spec: The data specification to write to
        """
        spec.comment("\nReserving memory space for data regions:\n\n")

        # Reserve memory:
        spec.reserve_memory_region(
            self._SYSTEM_REGION, SIMULATION_N_BYTES, label='setup')

        spec.reserve_memory_region(
            self._PARAMS_REGION, self._PARAMS_SIZE, label='params')

        self.reserve_provenance_data_region(spec)

    @overrides(MachineVertex.get_n_keys_for_partition)
    def get_n_keys_for_partition(self, partition_id: str) -> int:
        if partition_id == self.MOTOR_PARTITION_ID:
            return self._MOTOR_N_KEYS
        return super().get_n_keys_for_partition(partition_id)



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Dict, Mapping, Sequence
from spinn_utilities.overrides import overrides
from spinn_front_end_common.interface.ds import DataType
from spinn_front_end_common.utilities.constants import BYTES_PER_WORD
from spynnaker.pyNN.data import SpynnakerDataView
from spynnaker.pyNN.exceptions import SpynnakerException
from .abstract_current_source import (
    AbstractCurrentSource, CurrentSourceIDs, CurrentParameter)


class StepCurrentSource(AbstractCurrentSource):
    """
    Current source where the amplitude changes based on a time array.
    """
    __slots__ = (
        "__amplitudes",
        "__times",
        "__parameters",
        "__parameter_types")

    def __init__(
            self, times: Sequence[int] = (), amplitudes: Sequence[float] = ()):
        """
        :param times:
        :param amplitudes:
        """
        # There's probably no need to actually store these as you can't
        # access them directly in pynn anyway
        time_convert_ms = SpynnakerDataView.get_simulation_time_step_per_ms()
        self.__times = [
            int(times[i] * time_convert_ms) for i in range(len(times))]
        self.__amplitudes = amplitudes

        if len(times) != len(amplitudes):
            raise SpynnakerException(
                f"In StepCurrentSource, len(times) is {len(times)}, "
                f" but len(amplitudes) is {len(amplitudes)}")

        self.__parameter_types = {
            'times': DataType.UINT32,  # arrays?
            'amplitudes': DataType.S1615}

        self.__parameters: Dict[str, CurrentParameter] = {
            'times': self.__times,
            'amplitudes': self.__amplitudes}

        super().__init__()

    @overrides(AbstractCurrentSource.set_parameters)
    def set_parameters(self, **parameters: CurrentParameter) -> None:
        for key, value in parameters.items():
            if key not in self.__parameters:
                # throw an exception
                raise SpynnakerException(f"{key} is not a parameter of {self}")
            if not isinstance(value, Sequence):
                raise TypeError
            if key == 'times':
                time_convert_ms = SpynnakerDataView.\
                    get_simulation_time_step_per_ms()
                self.__times = [
                    int(value[i] * time_convert_ms) for i in range(len(value))]
                value = self.__times
            else:
                # Check length: if longer, need to remap
                if len(self.__amplitudes) < len(value):
                    if self.population is not None:
                        SpynnakerDataView.set_requires_mapping()

                self.__amplitudes = value
            self.__parameters[key] = value

        # Check the arrays are still the same lengths
        if len(self.__times) != len(self.__amplitudes):
            raise SpynnakerException(
                f"In StepCurrentSource, len(times) is {len(self.__times)}, "
                f"but len(amplitudes) is {len(self.__amplitudes)}")

        # Parameters have been set, so if multi-run then it will have been
        # injected already; if not then it can just be ignored
        if self.app_vertex is not None:
            for m_vertex in self.app_vertex.machine_vertices:
                m_vertex.set_reload_required(True)

    @property
    @overrides(AbstractCurrentSource.parameters)
    def parameters(self) -> Mapping[str, CurrentParameter]:
        return self.__parameters

    @property
    @overrides(AbstractCurrentSource.parameter_types)
    def parameter_types(self) -> Mapping[str, DataType]:
        return self.__parameter_types

    @property
    @overrides(AbstractCurrentSource.current_source_id)
    def current_source_id(self) -> int:
        return CurrentSourceIDs.STEP_CURRENT_SOURCE.value

    @overrides(AbstractCurrentSource.get_sdram_usage_in_bytes)
    def get_sdram_usage_in_bytes(self) -> int:
        # The parameters themselves take up this amount of space
        # length(times) + length_val) * 2) + ID
        sdram_for_parameters = ((len(self.__times) + 1) * 2) * BYTES_PER_WORD

        # For each_source there is the last amplitude holder and index
        sdram_for_on_core_calcs = 2 * BYTES_PER_WORD

        return sdram_for_parameters + sdram_for_on_core_calcs
        # return sdram_for_parameters



# Copyright (c) 2015 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from spinn_utilities.abstract_base import AbstractBase
from spynnaker.pyNN.models.neuron.implementations import (
    AbstractStandardNeuronComponent)


class AbstractAdditionalInput(  # pylint: disable=abstract-method
        AbstractStandardNeuronComponent, metaclass=AbstractBase):
    """
    Represents a possible additional independent input for a model.
    """

    __slots__ = ()



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Any, Mapping, Optional, Sequence

from spinn_utilities.overrides import overrides
from spinn_utilities.ranged.range_dictionary import RangeDictionary

from spinn_front_end_common.interface.ds import DataType
from spinn_front_end_common.utilities.constants import BYTES_PER_WORD

from spynnaker.pyNN.models.neuron.additional_inputs import (
    AbstractAdditionalInput)
from spynnaker.pyNN.models.neuron.input_types import (
    AbstractInputType, InputTypeConductance)
from spynnaker.pyNN.models.neuron.neuron_models import NeuronModel
from spynnaker.pyNN.models.neuron.synapse_types import AbstractSynapseType
from spynnaker.pyNN.models.neuron.threshold_types import AbstractThresholdType
from spynnaker.pyNN.utilities.struct import Struct, StructRepeat

from .abstract_neuron_impl import AbstractNeuronImpl

# The size of the n_steps_per_timestep parameter
_N_STEPS_PER_TIMESTEP_SIZE = 1 * BYTES_PER_WORD

# The default number of steps per timestep
_DEFAULT_N_STEPS_PER_TIMESTEP = 1

_STEPS_PER_TIMESTEP = "n_steps_per_timestep"
_STEPS_PER_TIMESTEP_STRUCT = Struct(
    [(DataType.UINT32, _STEPS_PER_TIMESTEP)], repeat_type=StructRepeat.GLOBAL)


class NeuronImplStandard(AbstractNeuronImpl):
    """
    The standard componentised neuron implementation.
    """

    __slots__ = (
        "__model_name",
        "__binary",
        "__neuron_model",
        "__input_type",
        "__synapse_type",
        "__threshold_type",
        "__additional_input_type",
        "__components",
        "__n_steps_per_timestep")

    _RECORDABLES = ["v", "gsyn_exc", "gsyn_inh"]

    _RECORDABLE_DATA_TYPES = {
        "v": DataType.S1615,
        "gsyn_exc": DataType.S1615,
        "gsyn_inh": DataType.S1615
    }

    _RECORDABLE_UNITS = {
        'v': 'mV',
        'gsyn_exc': "uS",
        'gsyn_inh': "uS"}

    def __init__(
            self, model_name: str, binary: str,
            neuron_model: NeuronModel, input_type: AbstractInputType,
            synapse_type: AbstractSynapseType,
            threshold_type: AbstractThresholdType,
            additional_input_type: Optional[AbstractAdditionalInput] = None):
        """
        :param model_name:
        :param binary:
        :param neuron_model:
        :param input_type:
        :param synapse_type:
        :param threshold_type:
        :param additional_input_type:
        """
        self.__model_name = model_name
        self.__binary = binary
        self.__neuron_model = neuron_model
        self.__input_type = input_type
        self.__synapse_type = synapse_type
        self.__threshold_type = threshold_type
        self.__additional_input_type = additional_input_type
        self.__n_steps_per_timestep = _DEFAULT_N_STEPS_PER_TIMESTEP

        self.__components = [
            self.__neuron_model, self.__input_type, self.__threshold_type,
            self.__synapse_type]
        if self.__additional_input_type is not None:
            self.__components.append(self.__additional_input_type)

    @property
    def n_steps_per_timestep(self) -> int:
        """
        Get the last set n steps per timestep
        """
        return self.__n_steps_per_timestep

    @n_steps_per_timestep.setter
    def n_steps_per_timestep(self, n_steps_per_timestep: int) -> None:
        self.__n_steps_per_timestep = n_steps_per_timestep

    @property
    @overrides(AbstractNeuronImpl.model_name)
    def model_name(self) -> str:
        return self.__model_name

    @property
    @overrides(AbstractNeuronImpl.binary_name)
    def binary_name(self) -> str:
        return self.__binary

    @property
    @overrides(AbstractNeuronImpl.structs)
    def structs(self) -> Sequence[Struct]:
        structs = [_STEPS_PER_TIMESTEP_STRUCT]
        structs.extend(s for c in self.__components for s in c.structs)
        return structs

    @overrides(AbstractNeuronImpl.get_global_weight_scale)
    def get_global_weight_scale(self) -> float:
        return self.__input_type.get_global_weight_scale()

    @overrides(AbstractNeuronImpl.get_n_synapse_types)
    def get_n_synapse_types(self) -> int:
        return self.__synapse_type.get_n_synapse_types()

    @overrides(AbstractNeuronImpl.get_synapse_id_by_target)
    def get_synapse_id_by_target(self, target: str) -> Optional[int]:
        return self.__synapse_type.get_synapse_id_by_target(target)

    @overrides(AbstractNeuronImpl.get_synapse_targets)
    def get_synapse_targets(self) -> Sequence[str]:
        return self.__synapse_type.get_synapse_targets()

    @overrides(AbstractNeuronImpl.get_recordable_variables)
    def get_recordable_variables(self) -> Sequence[str]:
        return self._RECORDABLES

    @overrides(AbstractNeuronImpl.get_recordable_units)
    def get_recordable_units(self, variable: str) -> str:
        return self._RECORDABLE_UNITS[variable]

    @overrides(AbstractNeuronImpl.get_recordable_data_types)
    def get_recordable_data_types(self) -> Mapping[str, DataType]:
        return self._RECORDABLE_DATA_TYPES

    @overrides(AbstractNeuronImpl.is_recordable)
    def is_recordable(self, variable: str) -> bool:
        return variable in self._RECORDABLES

    @overrides(AbstractNeuronImpl.get_recordable_variable_index)
    def get_recordable_variable_index(self, variable: str) -> int:
        return self._RECORDABLES.index(variable)

    @overrides(AbstractNeuronImpl.add_parameters)
    def add_parameters(self, parameters: RangeDictionary) -> None:
        parameters[_STEPS_PER_TIMESTEP] = self.__n_steps_per_timestep
        for component in self.__components:
            component.add_parameters(parameters)

    @overrides(AbstractNeuronImpl.add_state_variables)
    def add_state_variables(self, state_variables: RangeDictionary) -> None:
        for component in self.__components:
            component.add_state_variables(state_variables)

    @overrides(AbstractNeuronImpl.get_units)
    def get_units(self, variable: str) -> str:
        for component in self.__components:
            if component.has_variable(variable):
                return component.get_units(variable)

        raise KeyError(
            f"The parameter {variable} does not exist in this neuron model")

    @property
    @overrides(AbstractNeuronImpl.is_conductance_based)
    def is_conductance_based(self) -> bool:
        return isinstance(self.__input_type, InputTypeConductance)

    def __getitem__(self, key: str) -> Any:
        # Find the property in the components...
        for component in self.__components:
            if hasattr(component, key):
                return getattr(component, key)
        # ... or fail
        raise AttributeError(
            f"'{self.__class__.__name__}' object has no attribute {key}")



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Iterable, Optional, Union

from spinn_utilities.socket_address import SocketAddress
from spinn_utilities.config_holder import (get_config_int, get_config_str)

from spinnman.messages.eieio import EIEIOPrefix, EIEIOType

from pacman.model.graphs.application import (
    ApplicationEdge, ApplicationVertex)

from spinn_front_end_common.utility_models import (
    ReverseIpTagMultiCastSource)
from spinn_front_end_common.utilities.utility_objs import (
    LivePacketGatherParameters)

from spynnaker.pyNN.data import SpynnakerDataView
from spynnaker.pyNN.models.populations import Population
from spynnaker.pyNN.models.spike_source.spike_source_poisson_vertex import (
    SpikeSourcePoissonVertex)
from spynnaker.pyNN.utilities.constants import (
    LIVE_POISSON_CONTROL_PARTITION_ID, SPIKE_PARTITION_ID)


class SpynnakerExternalDevicePluginManager(object):
    """
    User-level interface for the external device plug-in manager.
    """
    __slots__ = ()

    @staticmethod
    def add_database_socket_address(
            database_notify_host: Optional[str],
            database_notify_port_num: Optional[int],
            database_ack_port_num: Optional[int]) -> None:
        """
        Creates a new SocketAddress and saves in into the data view

        :param database_notify_host:
            Host to talk to tell that the database (and application) is ready.
        :param database_notify_port_num:
            Port to talk to tell that the database (and application) is ready.
        :param database_ack_port_num:
            Port on which to listen for an acknowledgement that the
            simulation should start.
        """
        # build the database socket address used by the notification interface
        database_socket = SocketAddress(
            listen_port=database_ack_port_num,
            notify_host_name=database_notify_host,
            notify_port_no=database_notify_port_num)

        # update socket interface with new demands.
        SpynnakerDataView.add_database_socket_address(database_socket)

    @staticmethod
    def activate_live_output_for(
            population: Population, *,
            database_notify_host: Optional[str] = None,
            database_notify_port_num: Optional[int] = None,
            database_ack_port_num: Optional[int] = None,
            port: Optional[int] = None, host: Optional[str] = None,
            tag: Optional[int] = None, strip_sdp: bool = True,
            use_prefix: bool = False, key_prefix: Optional[int] = None,
            prefix_type: Optional[EIEIOPrefix] = None,
            message_type: EIEIOType = EIEIOType.KEY_32_BIT,
            right_shift: int = 0, payload_as_time_stamps: bool = True,
            notify: bool = True, use_payload_prefix: bool = True,
            payload_prefix: Optional[int] = None, payload_right_shift: int = 0,
            number_of_packets_sent_per_time_step: int = 0,
            translate_keys: bool = False,
            partition_ids: Optional[Iterable[str]] = None) -> None:
        """
        Output the spikes from a given population from SpiNNaker as they
        occur in the simulation.

        :param population: The population to activate the live output for
        :param database_notify_host:
            The hostname for the device which is listening to the database
            notification.
        :param database_ack_port_num:
            The port number to which a external device will acknowledge that
            they have finished reading the database and are ready for it to
            start execution
        :param database_notify_port_num:
            The port number to which a external device will receive the
            database is ready command
        :param key_prefix: the prefix to be applied to the key
        :param prefix_type:
            if the prefix type is 32 bit or 16 bit
        :param message_type:
            If the message is a EIEIO command message, or an EIEIO data
            message with 16 bit or 32 bit keys.
        :param payload_as_time_stamps:
        :param right_shift:
        :param use_payload_prefix:
        :param notify:
        :param payload_prefix:
        :param payload_right_shift:
        :param number_of_packets_sent_per_time_step:
        :param port:
            The UDP port to which the live spikes will be sent. If not
            specified, the port will be taken from the "live_spike_port"
            parameter in the "Recording" section of the sPyNNaker
            configuration file.
        :param host:
            The host name or IP address to which the live spikes will be
            sent. If not specified, the host will be taken from the
            "live_spike_host" parameter in the "Recording" section of the
            sPyNNaker configuration file.
        :param tag:
            The IP tag to be used for the spikes. If not specified, one will
            be automatically assigned
        :param strip_sdp:
            Determines if the SDP headers will be stripped from the
            transmitted packet.
        :param use_prefix:
            Determines if the spike packet will contain a common prefix for
            the spikes
        :param partition_ids:
            The names of the partitions to create edges for
        :param translate_keys:
            Whether the incoming keys from the cores should be translated
            to global keys rather than core-based keys
        """
        # pylint: disable=protected-access
        # get default params if none set
        if port is None:
            port = get_config_int("Recording", "live_spike_port")
        if host is None:
            host = get_config_str("Recording", "live_spike_host")

        # Use the right-shift to remove the colour from translated keys
        n_colour_bits = population._vertex.n_colour_bits
        translated_key_right_shift = n_colour_bits

        # Use the mask to remove the colour from non-translated keys
        received_key_mask = 0xFFFFFFFF & ~((2 ** n_colour_bits) - 1)

        if partition_ids is None:
            partition_ids = [SPIKE_PARTITION_ID]

        params = LivePacketGatherParameters(
            port=port, hostname=host, tag=tag, strip_sdp=strip_sdp,
            use_prefix=use_prefix, key_prefix=key_prefix,
            prefix_type=prefix_type, message_type=message_type,
            right_shift=right_shift, payload_prefix=payload_prefix,
            payload_as_time_stamps=payload_as_time_stamps,
            use_payload_prefix=use_payload_prefix,
            payload_right_shift=payload_right_shift,
            number_of_packets_sent_per_time_step=(
                number_of_packets_sent_per_time_step),
            label="LiveSpikeReceiver", received_key_mask=received_key_mask,
            translate_keys=translate_keys,
            translated_key_right_shift=translated_key_right_shift)
        SpynnakerExternalDevicePluginManager.update_live_packet_gather_tracker(
            population._vertex, params, partition_ids)

        if notify:
            SpynnakerExternalDevicePluginManager.add_database_socket_address(
                database_notify_host, database_notify_port_num,
                database_ack_port_num)

    @staticmethod
    def activate_live_output_to(
            population: Population,
            device: Union[Population, ApplicationVertex],
            partition_id: str = SPIKE_PARTITION_ID) -> None:
        """
        Activate the output of spikes from a population to an external device.

        .. note::
            All spikes will be sent to the device.

        :param population:
            The pyNN population object from which spikes will be sent.
        :param device:
            The pyNN population or external device to which the spikes will be
            sent.
        :param partition_id:
            The partition ID to activate live output to.
        """
        # pylint: disable=protected-access
        if isinstance(device, Population):
            device_vertex: ApplicationVertex = device._vertex
        else:
            device_vertex = device
        SpynnakerExternalDevicePluginManager.add_edge(
            population._vertex, device_vertex, partition_id)

    @staticmethod
    def update_live_packet_gather_tracker(
            vertex_to_record_from: ApplicationVertex,
            params: LivePacketGatherParameters,
            partition_ids: Iterable[str]) -> None:
        """
        Add an edge from a vertex to the live packet gatherer, builds as
        needed and has all the parameters for the creation of the live
        packet gatherer if needed.

        :param vertex_to_record_from:
        :param params:
        :param partition_ids:
        """
        # add to the tracker
        SpynnakerDataView.add_live_packet_gatherer_parameters(
            params, vertex_to_record_from, partition_ids)

    @staticmethod
    def add_poisson_live_rate_control(
            poisson_population: Population, *,
            control_label_extension: str = "_control",
            receive_port: Optional[int] = None,
            database_notify_host: Optional[str] = None,
            database_notify_port_num: Optional[int] = None,
            database_ack_port_num: Optional[int] = None, notify: bool = True,
            reserve_reverse_ip_tag: bool = False) -> None:
        """
        Add a live rate controller to a Poisson population.

        :param poisson_population: The population to control
        :param control_label_extension:
            An extension to add to the label of the Poisson source. Must match
            up with the equivalent in the SpynnakerPoissonControlConnection
        :param receive_port:
            The port that the SpiNNaker board should listen on
        :param database_notify_host: the hostname for the device which is
            listening to the database notification.
        :param database_ack_port_num: the port number to which a external
            device will acknowledge that they have finished reading the
            database and are ready for it to start execution
        :param database_notify_port_num: The port number to which an
            external device will receive the database is ready command
        :param notify: adds to the notification protocol if set.
        :param reserve_reverse_ip_tag: True if a reverse IP tag is to be
            used, False if SDP is to be used (default)
        """
        # pylint: disable=protected-access
        vertex = poisson_population._vertex
        if not isinstance(vertex, SpikeSourcePoissonVertex):
            raise TypeError("population must contain a SpikeSourcePoisson")
        control_label = f"{vertex.label}{control_label_extension}"
        controller = ReverseIpTagMultiCastSource(
            n_keys=vertex.n_atoms, label=control_label,
            receive_port=receive_port,
            reserve_reverse_ip_tag=reserve_reverse_ip_tag)
        SpynnakerExternalDevicePluginManager.add_application_vertex(controller)
        edge = SpynnakerExternalDevicePluginManager.add_edge(
            controller, vertex, LIVE_POISSON_CONTROL_PARTITION_ID)
        vertex.set_live_poisson_control_edge(edge)
        if notify:
            SpynnakerExternalDevicePluginManager.add_database_socket_address(
                database_notify_host, database_notify_port_num,
                database_ack_port_num)

    @staticmethod
    def add_edge(vertex: ApplicationVertex, device_vertex: ApplicationVertex,
                 partition_id: str) -> ApplicationEdge:
        """
        Add an edge between two vertices (often a vertex and a external
        device) on a given partition.

        :param vertex:
            the pre-population vertex to connect the edge from
        :param device_vertex:
            the post-population vertex to connect the edge to
        :param partition_id: the partition identifier for making nets
        :returns: Edge added
        """
        edge = ApplicationEdge(vertex, device_vertex)
        SpynnakerDataView.add_edge(edge, partition_id)
        return edge

    @staticmethod
    def add_application_vertex(vertex: ApplicationVertex) -> None:
        """
        Adds an Application vertex to the user graph.

        Semantic sugar for SpynnakerDataView.add_vertex(vertex)

        :param vertex:
        """
        SpynnakerDataView.add_vertex(vertex)



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Iterable, List, Optional
from spinn_utilities.overrides import overrides
from pacman.model.routing_info import BaseKeyAndMask
from pacman.model.graphs.application import ApplicationSpiNNakerLinkVertex
from spinn_front_end_common.abstract_models import (
    AbstractSendMeMulticastCommandsVertex)
from spinn_front_end_common.utility_models import MultiCastCommand
from spynnaker.pyNN.exceptions import SpynnakerException
from spynnaker.pyNN.models.common import PopulationApplicationVertex


def get_y_from_fpga_retina(key: int, mode: int) -> Optional[int]:
    """
    :param key:
    :param mode:
    :returns: y value if mode is expected otherwise None.
    """
    if mode == 128:
        return key & 0x7f
    elif mode == 64:
        return key & 0x3f
    elif mode == 32:
        return key & 0x1f
    elif mode == 16:
        return key & 0xf
    return None


def get_x_from_fpga_retina(key: int, mode: int) -> Optional[int]:
    """
    :param key:
    :param mode:
    :returns: x value if mode is expected otherwise None.
    """
    if mode == 128:
        return (key >> 7) & 0x7f
    elif mode == 64:
        return (key >> 6) & 0x3f
    elif mode == 32:
        return (key >> 5) & 0x1f
    elif mode == 16:
        return (key >> 4) & 0xf
    return None


def get_spike_value_from_fpga_retina(key: int, mode: int) -> Optional[int]:
    """
    :param key:
    :param mode:
    :returns: Spike value if mode is expected otherwise None.
    """
    if mode == 128:
        return (key >> 14) & 0x1
    elif mode == 64:
        return (key >> 14) & 0x1
    elif mode == 32:
        return (key >> 14) & 0x1
    elif mode == 16:
        return (key >> 14) & 0x1
    return None


class ExternalFPGARetinaDevice(
        ApplicationSpiNNakerLinkVertex, PopulationApplicationVertex,
        AbstractSendMeMulticastCommandsVertex):
    """
    A retina connected by FPGA
    """
    __slots__ = (
        "__fixed_key",
        "__fixed_mask")

    MODE_128 = "128"
    MODE_64 = "64"
    MODE_32 = "32"
    MODE_16 = "16"
    UP_POLARITY = "UP"
    DOWN_POLARITY = "DOWN"
    MERGED_POLARITY = "MERGED"

    def __init__(
            self, mode: str, retina_key: int, spinnaker_link_id: int,
            polarity: str, label: Optional[str] = None,
            board_address: Optional[str] = None):
        """
        :param mode: The retina "mode"
        :param retina_key: The value of the top 16-bits of the key
        :param spinnaker_link_id:
            The SpiNNaker link to which the retina is connected
        :param polarity: The "polarity" of the retina data
        :param label: The optional name of the vertex.
        :param board_address:
        """
        fixed_n_neurons = self.get_n_neurons(mode, polarity)
        super().__init__(
            n_atoms=fixed_n_neurons, spinnaker_link_id=spinnaker_link_id,
            label=label, board_address=board_address, incoming=True,
            outgoing=True)

        self.__fixed_key = (retina_key & 0xFFFF) << 16
        self.__fixed_mask = 0xFFFF8000
        if polarity == self.UP_POLARITY:
            self.__fixed_key |= 0x4000

        self.__fixed_mask = self._get_mask(mode)

    @overrides(ApplicationSpiNNakerLinkVertex.get_fixed_key_and_mask)
    def get_fixed_key_and_mask(self, partition_id: str) -> BaseKeyAndMask:
        return BaseKeyAndMask(self.__fixed_key, self.__fixed_mask)

    def _get_mask(self, mode: str) -> int:
        if mode == ExternalFPGARetinaDevice.MODE_128:
            return 0xFFFFC000
        elif mode == ExternalFPGARetinaDevice.MODE_64:
            return 0xFFFFF000
        elif mode == ExternalFPGARetinaDevice.MODE_32:
            return 0xFFFFFC00
        elif mode == ExternalFPGARetinaDevice.MODE_16:
            return 0xFFFFFF00
        raise SpynnakerException(
            "the FPGA retina does not recognise this mode")

    @staticmethod
    def get_n_neurons(mode: str, polarity: str) -> int:
        """
        Calculate the number of neurons based on mode and polarity

        :param mode: ``128`` or ``64`` or ``32`` or ``16``
        :param polarity: ``UP`` or ``DOWN``
        :returns: The number of neurons in the retina
        """
        # Uses mode and polarity to determine the number of neurons needed
        if mode == ExternalFPGARetinaDevice.MODE_128:
            if (polarity == ExternalFPGARetinaDevice.UP_POLARITY or
                    polarity == ExternalFPGARetinaDevice.DOWN_POLARITY):
                return 128 * 128
            return 128 * 128 * 2
        elif mode == ExternalFPGARetinaDevice.MODE_64:
            if (polarity == ExternalFPGARetinaDevice.UP_POLARITY or
                    polarity == ExternalFPGARetinaDevice.DOWN_POLARITY):
                return 64 * 64
            return 64 * 64 * 2
        elif mode == ExternalFPGARetinaDevice.MODE_32:
            if (polarity == ExternalFPGARetinaDevice.UP_POLARITY or
                    polarity == ExternalFPGARetinaDevice.DOWN_POLARITY):
                return 32 * 32
            return 32 * 32 * 2
        elif mode == ExternalFPGARetinaDevice.MODE_16:
            if (polarity == ExternalFPGARetinaDevice.UP_POLARITY or
                    polarity == ExternalFPGARetinaDevice.DOWN_POLARITY):
                return 16 * 16
            return 16 * 16 * 2
        raise SpynnakerException(
            "the FPGA retina does not recognise this mode")

    @property
    @overrides(AbstractSendMeMulticastCommandsVertex.start_resume_commands)
    def start_resume_commands(self) -> Iterable[MultiCastCommand]:
        yield MultiCastCommand(
            key=0x0000FFFF, payload=1, repeat=5, delay_between_repeats=100)

    @property
    @overrides(AbstractSendMeMulticastCommandsVertex.pause_stop_commands)
    def pause_stop_commands(self) -> Iterable[MultiCastCommand]:
        yield MultiCastCommand(
            key=0x0000FFFE, payload=0, repeat=5, delay_between_repeats=100)

    @property
    @overrides(AbstractSendMeMulticastCommandsVertex.timed_commands)
    def timed_commands(self) -> List[MultiCastCommand]:
        return []



# Copyright (c) 2020 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import List, Sequence, Tuple, Union

import numpy
from numpy.typing import NDArray
from typing_extensions import TypeAlias

from spynnaker.pyNN.random_distribution import RandomDistribution

#: Type of names of parameters and state variables.
Names: TypeAlias = Union[str, List[str], Tuple[str, ...]]

#: Type of normal values of parameters and state variables.
Values: TypeAlias = Union[
    float, Sequence[float], NDArray[numpy.floating], RandomDistribution]

#: Type of spikes in spike sources.
Spikes: TypeAlias = Union[
    # Can be floating point values (will round)
    Values,
    # Can be integer values, or lists of such
    int, Sequence[int], Sequence[Sequence[int]], NDArray[numpy.integer]]



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Iterable
from spinn_utilities.overrides import overrides
from spinn_front_end_common.interface.ds import DataSpecificationBase
from .abstract_partner_selection import AbstractPartnerSelection


class LastNeuronSelection(AbstractPartnerSelection):
    """
    Partner selection that picks a random source neuron from the neurons
    that spiked in the last timestep.
    """

    __slots__ = ("__spike_buffer_size", )

    def __init__(self, spike_buffer_size: int = 64):
        """
        :param spike_buffer_size: The size of the buffer for holding spikes
        """
        self.__spike_buffer_size = spike_buffer_size

    @property
    @overrides(AbstractPartnerSelection.vertex_executable_suffix)
    def vertex_executable_suffix(self) -> str:
        return "_last_neuron"

    @overrides(AbstractPartnerSelection.get_parameters_sdram_usage_in_bytes)
    def get_parameters_sdram_usage_in_bytes(self) -> int:
        return 4

    @overrides(AbstractPartnerSelection.write_parameters)
    def write_parameters(self, spec: DataSpecificationBase) -> None:
        spec.write_value(self.__spike_buffer_size)

    @overrides(AbstractPartnerSelection.get_parameter_names)
    def get_parameter_names(self) -> Iterable[str]:
        yield "spike_buffer_size"



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from spinn_utilities.overrides import overrides
from spinn_utilities.ranged import RangeDictionary

from spinn_front_end_common.interface.ds import DataType

from spynnaker.pyNN.models.neuron.implementations import ModelParameter
from spynnaker.pyNN.utilities.struct import Struct
from spynnaker.pyNN.data import SpynnakerDataView

from .abstract_threshold_type import AbstractThresholdType

DU_TH = "du_th"
TAU_TH = "tau_th"
V_THRESH = "v_thresh"
TIMESTEP = "timestep"


class ThresholdTypeMaassStochastic(AbstractThresholdType):
    """
    A stochastic threshold.

    Habenschuss S, Jonke Z, Maass W. Stochastic computations in cortical
    microcircuit models. *PLoS Computational Biology.* 2013;9(11):e1003311.
    `doi:10.1371/journal.pcbi.1003311
    <https://doi.org/10.1371/journal.pcbi.1003311>`_
    """
    __slots__ = (
        "__du_th",
        "__tau_th",
        "__v_thresh")

    def __init__(self, du_th: ModelParameter, tau_th: ModelParameter,
                 v_thresh: ModelParameter):
        r"""
        :param du_th: :math:`du_{thresh}`
        :param tau_th: :math:`\tau_{thresh}`
        :param v_thresh: :math:`V_{thresh}`
        """
        super().__init__(
            [Struct([
                (DataType.S1615, DU_TH),
                (DataType.S1615, TAU_TH),
                (DataType.S1615, V_THRESH),
                (DataType.S1615, TIMESTEP)])],
            {DU_TH: "mV", TAU_TH: "ms", V_THRESH: "mV"})
        self.__du_th = du_th
        self.__tau_th = tau_th
        self.__v_thresh = v_thresh

    @overrides(AbstractThresholdType.add_parameters)
    def add_parameters(self, parameters: RangeDictionary[float]) -> None:
        parameters[DU_TH] = self._convert(self.__du_th)
        parameters[TAU_TH] = self._convert(self.__tau_th)
        parameters[V_THRESH] = self._convert(self.__v_thresh)
        parameters[TIMESTEP] = SpynnakerDataView.get_simulation_time_step_ms()

    @overrides(AbstractThresholdType.add_state_variables)
    def add_state_variables(
            self, state_variables: RangeDictionary[float]) -> None:
        pass

    @property
    def v_thresh(self) -> ModelParameter:
        """
        :math:`V_{thresh}`
        """
        return self.__v_thresh

    @property
    def du_th(self) -> ModelParameter:
        """
        :math:`du_{thresh}`
        """
        return self.__du_th

    @property
    def tau_th(self) -> ModelParameter:
        r"""
        :math:`\tau_{thresh}`
        """
        return self.__tau_th



# Copyright (c) 2024 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations
from typing import Iterable, List, Optional, Tuple, TYPE_CHECKING, cast

import numpy
from numpy import floating, integer, uint8, uint32, int16
from numpy.typing import NDArray

from spinn_utilities.overrides import overrides
from pacman.utilities.utility_calls import get_n_bits

from spinn_front_end_common.interface.ds import DataSpecificationBase
from spinn_front_end_common.utilities.constants import BYTES_PER_WORD

from spynnaker.pyNN.exceptions import (
    SynapticConfigurationException, InvalidParameterType)
from spynnaker.pyNN.models.neuron.synapse_dynamics.types import (
    NUMPY_CONNECTORS_DTYPE)
from .abstract_plastic_synapse_dynamics import AbstractPlasticSynapseDynamics
from .abstract_generate_on_machine import AbstractGenerateOnMachine
from .abstract_generate_on_machine import MatrixGeneratorID
from .synapse_dynamics_weight_changable import SynapseDynamicsWeightChangable

if TYPE_CHECKING:
    from spynnaker.pyNN.models.neural_projections import (
        ProjectionApplicationEdge, SynapseInformation)
    from spynnaker.pyNN.models.neuron.synapse_io import MaxRowInfo
    from spynnaker.pyNN.models.neuron.synapse_dynamics.types import (
        ConnectionsArray)
    from spynnaker.pyNN.models.neural_projections.connectors import (
        AbstractConnector)
    from spynnaker.pyNN.types import WEIGHTS
    from spynnaker.pyNN.models.projection import Projection
    from .abstract_synapse_dynamics import AbstractSynapseDynamics


class SynapseDynamicsWeightChanger(
        AbstractPlasticSynapseDynamics, AbstractGenerateOnMachine):
    """
    Synapses that target a weight change
    """

    __slots__ = ["__post_vertex", "__synapse_info"]

    def __init__(self, weight_change: float, projection: Projection):
        """
        :param weight_change:
            The positive or negative change in weight to apply on each spike
        :param projection:
            The projection that this synapse dynamics is being added to
        """
        super().__init__(delay=1, weight=weight_change)
        # pylint: disable=protected-access
        self.__synapse_info = projection._synapse_information
        if not isinstance(self.__synapse_info.synapse_dynamics,
                          SynapseDynamicsWeightChangable):
            raise SynapticConfigurationException(
                "A changer can only affect a changeable projection")
        # Note: we store the post vertex here rather than the dynamics, as the
        # dynamics can change over time
        # Import here required to avoid circular imports
        # pylint: disable=import-outside-toplevel
        from spynnaker.pyNN.models.neuron import PopulationVertex
        self.__post_vertex = cast(PopulationVertex,
                                  self.__synapse_info.post_vertex)

    @overrides(AbstractPlasticSynapseDynamics.merge)
    def merge(self, synapse_dynamics: AbstractSynapseDynamics
              ) -> AbstractSynapseDynamics:
        # This must replace something that supports weight change,
        # so it can't be the first thing to be merged!
        raise SynapticConfigurationException(
            "Weight Changer synapses can only be added where an existing"
            " projection has already been added which supports"
            " weight changes")

    @overrides(AbstractPlasticSynapseDynamics.is_same_as)
    def is_same_as(self, synapse_dynamics: AbstractSynapseDynamics) -> bool:
        # Shouldn't ever come up, but if it does, it is False!
        return False

    @overrides(AbstractPlasticSynapseDynamics.get_vertex_executable_suffix)
    def get_vertex_executable_suffix(self) -> str:
        return "_weight_change"

    @overrides(AbstractPlasticSynapseDynamics
               .get_parameters_sdram_usage_in_bytes)
    def get_parameters_sdram_usage_in_bytes(
            self, n_neurons: int, n_synapse_types: int) -> int:
        # Should never be asked!
        return 0

    @overrides(AbstractPlasticSynapseDynamics.write_parameters)
    def write_parameters(
            self, spec: DataSpecificationBase, region: int,
            global_weight_scale: float,
            synapse_weight_scales: NDArray[floating]) -> None:
        # Should never be asked!
        pass

    @overrides(AbstractPlasticSynapseDynamics.get_value)
    def get_value(self, key: str) -> float:
        if hasattr(self, key):
            return getattr(self, key)
        raise InvalidParameterType(
            f"Type {type(self)} does not have parameter {key}")

    @overrides(AbstractPlasticSynapseDynamics.set_value)
    def set_value(self, key: str, value: float) -> None:
        if hasattr(self, key):
            setattr(self, key, value)
        raise InvalidParameterType(
            f"Type {type(self)} does not have parameter {key}")

    @overrides(AbstractPlasticSynapseDynamics
               .get_n_words_for_plastic_connections)
    def get_n_words_for_plastic_connections(self, n_connections: int) -> int:
        # 1 for flag and pre-spike identifier
        pp_size_words = 1
        # 1 for each connection
        fp_size_words = n_connections
        return pp_size_words + fp_size_words

    @overrides(AbstractPlasticSynapseDynamics.get_plastic_synaptic_data)
    def get_plastic_synaptic_data(
            self, connections: ConnectionsArray,
            connection_row_indices: NDArray[integer], n_rows: int,
            n_synapse_types: int,
            max_n_synapses: int, max_atoms_per_core: int) -> Tuple[
                NDArray[uint32], NDArray[uint32], NDArray[uint32],
                NDArray[uint32]]:
        weights = numpy.rint(numpy.abs(connections["weight"]))
        n_neuron_id_bits = get_n_bits(max_atoms_per_core)
        neuron_id_mask = (1 << n_neuron_id_bits) - 1
        fixed_plastic = (
            ((weights.astype(uint32) & 0xFFFF) << 16) |
            (connections["synapse_type"].astype(uint32) << n_neuron_id_bits) |
            (connections["target"] & neuron_id_mask))
        fixed_plastic_rows = self.convert_per_connection_data_to_rows(
            connection_row_indices, n_rows,
            fixed_plastic.view(dtype=uint8).reshape((-1, BYTES_PER_WORD)),
            max_n_synapses)

        flags = 0x80000000 | numpy.arange(n_rows, dtype=uint32)

        fp_size = self.get_n_items(fixed_plastic_rows, BYTES_PER_WORD)
        fp_data = numpy.vstack([
            fixed_row.view(uint32) for fixed_row in fixed_plastic_rows])
        pp_data = flags.astype(uint32)
        pp_size = numpy.ones(n_rows, dtype=uint32)

        return fp_data, pp_data, fp_size, pp_size

    @overrides(
        AbstractPlasticSynapseDynamics.get_n_plastic_plastic_words_per_row)
    def get_n_plastic_plastic_words_per_row(
            self, pp_size: NDArray[integer]) -> NDArray[integer]:
        # pp_size is in words, so just return
        return pp_size

    @overrides(
        AbstractPlasticSynapseDynamics.get_n_fixed_plastic_words_per_row)
    def get_n_fixed_plastic_words_per_row(
            self, fp_size: NDArray[integer]) -> NDArray[integer]:
        # fp_size is in words, so just return
        return fp_size

    @overrides(AbstractPlasticSynapseDynamics.get_n_synapses_in_rows)
    def get_n_synapses_in_rows(
            self, pp_size: NDArray[integer],
            fp_size: NDArray[integer]) -> NDArray[integer]:
        # Each fixed-plastic synapse is a word and fp_size is in words so just
        # return it
        return fp_size

    @overrides(AbstractPlasticSynapseDynamics.read_plastic_synaptic_data)
    def read_plastic_synaptic_data(
            self, n_synapse_types: int,
            pp_size: NDArray[uint32], pp_data: List[NDArray[uint32]],
            fp_size: NDArray[uint32], fp_data: List[NDArray[uint32]],
            max_atoms_per_core: int) -> ConnectionsArray:
        data = numpy.concatenate(fp_data)
        weight = ((data >> 16) & 0xFFFF).astype(int16)
        n_neuron_id_bits = get_n_bits(max_atoms_per_core)
        neuron_id_mask = (1 << n_neuron_id_bits) - 1
        connections = numpy.zeros(data.size, dtype=NUMPY_CONNECTORS_DTYPE)
        connections["source"] = numpy.concatenate(
            [numpy.repeat(i, fp_size[i]) for i in range(len(fp_size))])
        connections["target"] = data & neuron_id_mask
        connections["weight"] = weight
        connections["delay"] = 1
        return connections

    @overrides(AbstractPlasticSynapseDynamics.get_parameter_names)
    def get_parameter_names(self) -> Iterable[str]:
        yield 'weight'

    @overrides(AbstractPlasticSynapseDynamics.get_max_synapses)
    def get_max_synapses(self, n_words: int) -> int:
        # One word is static, the rest is for synapses
        return n_words - 1

    @property
    @overrides(AbstractGenerateOnMachine.gen_matrix_id)
    def gen_matrix_id(self) -> int:
        return MatrixGeneratorID.CHANGE_WEIGHT_MATRIX.value

    @overrides(AbstractGenerateOnMachine.gen_matrix_params)
    def gen_matrix_params(
            self, synaptic_matrix_offset: int, delayed_matrix_offset: int,
            app_edge: ProjectionApplicationEdge,
            synapse_info: SynapseInformation, max_row_info: MaxRowInfo,
            max_pre_atoms_per_core: int,
            max_post_atoms_per_core: int) -> NDArray[uint32]:
        _ = (delayed_matrix_offset, synapse_info, max_pre_atoms_per_core)
        vertex = app_edge.post_vertex
        n_synapse_type_bits = get_n_bits(
            vertex.neuron_impl.get_n_synapse_types())
        n_synapse_index_bits = get_n_bits(max_post_atoms_per_core)
        # We need to use the "global" dynamics object to get the offset
        dynamics = cast(SynapseDynamicsWeightChangable,
                        app_edge.post_vertex.synapse_dynamics)
        row_offset = dynamics.get_synapse_info_index(self.__synapse_info)
        return numpy.array([
            synaptic_matrix_offset, max_row_info.undelayed_max_words,
            max_row_info.undelayed_max_n_synapses,
            app_edge.pre_vertex.n_atoms, self.__synapse_info.synapse_type,
            n_synapse_type_bits, n_synapse_index_bits, row_offset],
            dtype=uint32)

    @property
    @overrides(AbstractGenerateOnMachine.
               gen_matrix_params_size_in_bytes)
    def gen_matrix_params_size_in_bytes(self) -> int:
        return 8 * BYTES_PER_WORD

    @property
    @overrides(AbstractPlasticSynapseDynamics.changes_during_run)
    def changes_during_run(self) -> bool:
        return False

    @property
    @overrides(AbstractPlasticSynapseDynamics.pad_to_length)
    def pad_to_length(self) -> None:
        return None

    @overrides(AbstractPlasticSynapseDynamics.get_synapse_id_by_target)
    def get_synapse_id_by_target(self, target: str) -> Optional[int]:
        return 0

    @property
    @overrides(AbstractPlasticSynapseDynamics.is_combined_core_capable)
    def is_combined_core_capable(self) -> bool:
        return True

    @property
    @overrides(AbstractPlasticSynapseDynamics.is_split_core_capable)
    def is_split_core_capable(self) -> bool:
        return True

    @overrides(AbstractPlasticSynapseDynamics.get_weight_maximum)
    def get_weight_maximum(
            self, connector: AbstractConnector,
            synapse_info: SynapseInformation) -> float:
        return self.__post_vertex.synapse_dynamics.get_weight_maximum(
            self.__synapse_info.connector, self.__synapse_info)

    @overrides(AbstractPlasticSynapseDynamics.get_weight_mean)
    def get_weight_mean(
            self, connector: AbstractConnector,
            synapse_info: SynapseInformation) -> float:
        return self.get_weight_maximum(connector, synapse_info)

    @overrides(AbstractPlasticSynapseDynamics.get_weight_variance)
    def get_weight_variance(
            self, connector: AbstractConnector, weights: WEIGHTS,
            synapse_info: SynapseInformation) -> float:
        return 0.0

    @overrides(AbstractPlasticSynapseDynamics.validate_connection)
    def validate_connection(
            self, application_edge: ProjectionApplicationEdge,
            synapse_info: SynapseInformation) -> None:
        if (application_edge.pre_vertex.n_atoms !=
                self.__synapse_info.pre_vertex.n_atoms):
            raise SynapticConfigurationException(
                "The pre-Population of this projection must have the same"
                " number of atoms as the pre-Population of the projection"
                " whose weights are to be changed")
        if application_edge.post_vertex != self.__synapse_info.post_vertex:
            raise SynapticConfigurationException(
                "The post-Population of this projection must be the same as"
                " the post-Population of the projection whose weights are to"
                " be changed")
        if synapse_info.synapse_type != self.__synapse_info.synapse_type:
            raise SynapticConfigurationException(
                "The synapse type of the projection must be the same as the"
                " synapse type of the projection whose weights are to be"
                " changed")
        AbstractPlasticSynapseDynamics.validate_connection(
            self, application_edge, synapse_info)

    @property
    @overrides(AbstractPlasticSynapseDynamics.synapses_per_second)
    def synapses_per_second(self) -> int:
        # This should never end up being requested!
        raise NotImplementedError



# Copyright (c) 2014 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from spinn_front_end_common.utilities.exceptions import ConfigurationException


class SpynnakerException(Exception):
    """
    Superclass of all exceptions from the PyNN module.
    """


class MemReadException(SpynnakerException):
    """
    Raised when the PyNN front end fails to read a certain memory region.
    """


class FilterableException(SpynnakerException):
    """
    Raised when it is not possible to determine if an edge should be filtered.
    """


class SynapticConfigurationException(ConfigurationException):
    """
    Raised when the synaptic manager fails for some reason.
    """


class SynapticBlockGenerationException(ConfigurationException):
    """
    Raised when the synaptic manager fails to generate a synaptic block.
    """


class SynapticBlockReadException(ConfigurationException):
    """
    Raised when the synaptic manager fails to read a synaptic block or
    convert it into readable values.
    """


class SynapticMaxIncomingAtomsSupportException(ConfigurationException):
    """
    Raised when a synaptic sublist exceeds the max atoms possible to be
    supported.
    """


class DelayExtensionException(ConfigurationException):
    """
    Raised when a delay extension vertex fails.
    """


class SpynnakerSplitterConfigurationException(ConfigurationException):
    """
    Raised when a splitter configuration fails.
    """


class InvalidParameterType(SpynnakerException):
    """
    Raised when a parameter is not recognised.
    """


class SynapseRowTooBigException(SpynnakerException):
    """
    Raised when a synapse row is bigger than is allowed.
    """
    def __init__(self, max_size: int, message: str):
        """
        :param max_size: the maximum permitted size of row
        :param message: the exception message
        """
        super().__init__(message)
        self._max_size = max_size

    @property
    def max_size(self) -> int:
        """
        The maximum size allowed.
        """
        return self._max_size



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Dict, Mapping

import numpy

from spinn_utilities.overrides import overrides

from spinn_front_end_common.interface.ds import DataType
from spinn_front_end_common.utilities.constants import BYTES_PER_WORD

from spynnaker.pyNN.data import SpynnakerDataView
from spynnaker.pyNN.exceptions import SpynnakerException

from .abstract_current_source import (
    AbstractCurrentSource, CurrentSourceIDs, CurrentParameter)


class ACSource(AbstractCurrentSource):
    """
    AC current source (i.e. sine wave) turned on at "start" and off at
    "stop", given (y-)offset, amplitude, frequency and phase.
    """
    __slots__ = (
        "__start",
        "__stop",
        "__amplitude",
        "__offset",
        "__frequency",
        "__phase",
        "__parameters",
        "__parameter_types")

    def __init__(self, start: float = 0.0, stop: float = 0.0,
                 amplitude: float = 0.0, offset: float = 0.0,
                 frequency: float = 0.0, phase: float = 0.0) -> None:
        """
        :param start:
        :param stop:
        :param amplitude:
        :param offset:
        :param frequency:
        :param phase:
        """
        # There's probably no need to actually store these as you can't
        # access them directly in pynn anyway
        time_convert_ms = SpynnakerDataView.get_simulation_time_step_per_ms()
        self.__start = int(start * time_convert_ms)
        self.__stop = int(stop * time_convert_ms)
        self.__amplitude = amplitude
        self.__offset = offset
        self.__frequency = self._get_frequency(frequency)
        self.__phase = self._get_phase(phase)

        self.__parameter_types = {
            'start': DataType.UINT32,
            'stop': DataType.UINT32,
            'amplitude': DataType.S1615,
            'offset': DataType.S1615,
            'frequency': DataType.S1615,
            'phase': DataType.S1615}

        self.__parameters: Dict[str, CurrentParameter] = {
            'start': self.__start,
            'stop': self.__stop,
            'amplitude': self.__amplitude,
            'offset': self.__offset,
            'frequency': self.__frequency,
            'phase': self.__phase}

        super().__init__()

    @overrides(AbstractCurrentSource.set_parameters)
    def set_parameters(self, **parameters: CurrentParameter) -> None:
        for key, value in parameters.items():
            if key not in self.__parameters:
                # throw an exception
                raise SpynnakerException(f"{key} is not a parameter of {self}")
            if key == 'frequency':
                self.__parameters[key] = self._get_frequency(value)
            elif key == 'phase':
                self.__parameters[key] = self._get_phase(value)
            else:
                self.__parameters[key] = value

        # Parameters have been set, so if multi-run then it will have been
        # injected already; if not then it can just be ignored
        if self.app_vertex is not None:
            for m_vertex in self.app_vertex.machine_vertices:
                m_vertex.set_reload_required(True)

    @property
    @overrides(AbstractCurrentSource.parameters)
    def parameters(self) -> Mapping[str, CurrentParameter]:
        return self.__parameters

    @property
    @overrides(AbstractCurrentSource.parameter_types)
    def parameter_types(self) -> Mapping[str, DataType]:
        return self.__parameter_types

    @property
    @overrides(AbstractCurrentSource.current_source_id)
    def current_source_id(self) -> int:
        return CurrentSourceIDs.AC_SOURCE.value

    @overrides(AbstractCurrentSource.get_sdram_usage_in_bytes)
    def get_sdram_usage_in_bytes(self) -> int:
        return len(self.__parameters) * BYTES_PER_WORD

    def _get_frequency(self, frequency: CurrentParameter) -> float:
        """
        Convert frequency to radian-friendly value.
        """
        if not isinstance(frequency, (int, float)):
            raise TypeError
        # convert frequency and phase into radians, remembering that
        # frequency is given in Hz but we are using ms for timesteps
        return (frequency * 2 * numpy.pi) / 1000.0

    def _get_phase(self, phase: CurrentParameter) -> float:
        """
        Convert phase to radian-friendly value.
        """
        if not isinstance(phase, (int, float)):
            raise TypeError
        return phase * (numpy.pi / 180.0)



# Copyright (c) 2014 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Utility package containing simple helper functions.
"""
from collections.abc import Sized
import logging
import os
import math
from math import isnan
from typing import cast, List, Optional, Tuple, Union

import neo
from neo.io.baseio import BaseIO  # type: ignore[import]
import numpy
from numpy import uint32, float64, floating
from numpy.typing import NDArray
from pyNN.random import AbstractRNG, RandomDistribution
from scipy.stats import binom

from spinn_utilities.log import FormatAdapter
from spinn_utilities.safe_eval import SafeEval
from spinn_utilities.config_holder import get_config_bool
from spinn_utilities.logger_utils import warn_once

from spinn_front_end_common.interface.ds import DataType
from spinn_front_end_common.utilities.constants import (
    MICRO_TO_SECOND_CONVERSION)
from spinn_front_end_common.utilities.exceptions import ConfigurationException

from spynnaker.pyNN.utilities.random_stats import (
    RandomStatsExponentialImpl, RandomStatsGammaImpl, RandomStatsLogNormalImpl,
    RandomStatsNormalClippedImpl, RandomStatsNormalImpl,
    RandomStatsPoissonImpl, RandomStatsRandIntImpl, RandomStatsUniformImpl,
    RandomStatsVonmisesImpl, RandomStatsBinomialImpl,
    RandomStatsExponentialClippedImpl)
from spynnaker.pyNN.utilities.constants import WRITE_BANDWIDTH_BYTES_PER_SECOND

logger = FormatAdapter(logging.getLogger(__name__))

MAX_RATE = 2 ** 32 - 1  # To allow a unit32_t to be used to store the rate

BASE_RANDOM_FOR_MARS_64 = 0x80000000
CAP_RANDOM_FOR_MARS_64 = 0x7FFFFFFF
# in order are x, y, z, c
N_RANDOM_NUMBERS = 4
ARBITRARY_Y = 13031301
MARS_C_MAX = 698769068

STATS_BY_NAME = {
    'binomial': RandomStatsBinomialImpl(),
    'gamma': RandomStatsGammaImpl(),
    'exponential': RandomStatsExponentialImpl(),
    'exponential_clipped': RandomStatsExponentialClippedImpl(),
    'lognormal': RandomStatsLogNormalImpl(),
    'normal': RandomStatsNormalImpl(),
    'normal_clipped': RandomStatsNormalClippedImpl(),
    'normal_clipped_to_boundary': RandomStatsNormalClippedImpl(),
    'poisson': RandomStatsPoissonImpl(),
    'uniform': RandomStatsUniformImpl(),
    'randint': RandomStatsRandIntImpl(),
    'vonmises': RandomStatsVonmisesImpl()}


def check_directory_exists_and_create_if_not(filename: str) -> None:
    """
    Create a parent directory for a file if it doesn't exist.

    :param filename: The file whose parent directory is to be created
    """
    directory = os.path.dirname(filename)
    if directory != "" and not os.path.exists(directory):
        os.makedirs(directory)


def convert_param_to_numpy(
        param: Union[RandomDistribution, float, List[float], NDArray],
        no_atoms: int) -> NDArray[floating]:
    """
    Convert parameters into numpy arrays.

    :param param: the param to convert
    :param no_atoms: the number of atoms available for conversion of param
    :return: the converted param as an array of floats
    """
    # Deal with random distributions by generating values
    if isinstance(param, RandomDistribution):
        # numpy reduces a single valued array to a single value, so enforce
        # that it is an array
        param_value = param.next(n=no_atoms)
        if hasattr(param_value, '__iter__'):
            return numpy.array(param_value, dtype=float64)
        return numpy.array([param_value], dtype=float64)

    # Deal with a single value by exploding to multiple values
    if not hasattr(param, '__iter__'):
        return numpy.array([param] * no_atoms, dtype=float64)

    # Deal with multiple values, but not the correct number of them
    if len(cast(Sized, param)) != no_atoms:
        raise ConfigurationException(
            "The number of params does not equal with the number of atoms in"
            " the vertex")

    # Deal with the correct number of multiple values
    return numpy.array(param, dtype=float64)


def convert_to(value: float, data_type: DataType) -> uint32:
    """
    Convert a value to a given data type.

    :param value: The value to convert
    :param data_type: The data type to convert to
    :return: The converted data as a numpy data type
    """
    return numpy.round(data_type.encode_as_int(value)).astype(
        data_type.struct_encoding)


def read_in_data_from_file(
        file_path: str, min_atom: int, max_atom: int,
        min_time: float, max_time: float, extra: bool = False) -> NDArray:
    """
    Read in a file of data values where the values are in a format of::

        <time>\t<atom ID>\t<data value>

    :param file_path: absolute path to a file containing the data
    :param min_atom: min neuron ID to which neurons to read in
    :param max_atom: max neuron ID to which neurons to read in
    :param extra:
    :param min_time: min time slot to read neurons values of.
    :param max_time: max time slot to read neurons values of.
    :return: a numpy array of (time stamp, atom ID, data value)
    """
    times: List[float] = []
    atom_ids: List[int] = []
    data_items: List[float] = []
    evaluator = SafeEval()
    with open(file_path, 'r', encoding="utf-8") as f:
        for line in f.readlines():
            if line.startswith('#'):
                continue
            if extra:
                time_s, neuron_id_s, data_value_s, _extra = line.split("\t")
            else:
                time_s, neuron_id_s, data_value_s = line.split("\t")
            time = float(evaluator.eval(time_s))
            neuron_id = int(evaluator.eval(neuron_id_s))
            data_value = float(evaluator.eval(data_value_s))
            if (min_atom <= neuron_id < max_atom and
                    min_time <= time < max_time):
                times.append(time)
                atom_ids.append(neuron_id)
                data_items.append(data_value)
            else:
                print(f"failed to enter {neuron_id}:{time}")

    result = numpy.dstack((atom_ids, times, data_items))[0]
    return result[numpy.lexsort(result.T[1::-1])]


def read_spikes_from_file(
        file_path: str,
        min_atom: float = 0, max_atom: float = float('inf'),
        min_time: float = 0, max_time: float = float('inf'),
        split_value: str = "\t") -> NDArray[numpy.integer]:
    """
    Read spikes from a file formatted as::

        <time>\t<neuron ID>

    :param file_path: absolute path to a file containing spike values
    :param min_atom: min neuron ID to which neurons to read in
    :param max_atom: max neuron ID to which neurons to read in
    :param min_time: min time slot to read neurons values of.
    :param max_time: max time slot to read neurons values of.
    :param split_value: the pattern to split by
    :return:
        a numpy array with up to max_atom elements each of which is a list of
        spike times.
    """
    # For backward compatibility as previous version tested for None rather
    # than having default values
    if min_atom is None:
        min_atom = 0.0
    if max_atom is None:
        max_atom = float('inf')
    if min_time is None:
        min_time = 0.0
    if max_time is None:
        max_time = float('inf')

    data = []
    with open(file_path, 'r', encoding="utf-8") as f_source:
        read_data = f_source.readlines()

    evaluator = SafeEval()
    for line in read_data:
        if line.startswith('#'):
            continue
        values = line.split(split_value)
        time = float(evaluator.eval(values[0]))
        neuron_id = float(evaluator.eval(values[1]))
        if (min_atom <= neuron_id < max_atom and
                min_time <= time < max_time):
            data.append([neuron_id, time])
    data.sort()
    return numpy.array(data)


def get_probable_maximum_selected(
        n_total_trials: int, n_trials: int, selection_prob: float,
        chance: float = (1.0 / 100.0)) -> int:
    """
    Get the likely maximum number of items that will be selected from a
    set of `n_trials` from a total set of `n_total_trials`
    with a probability of selection of `selection_prob`.

    :returns: The likely maximum number of items
    """
    prob = 1.0 - (chance / float(n_total_trials))
    val = binom.ppf(prob, n_trials, selection_prob)
    if isnan(val):
        raise ValueError(
            f"Could not find maximum selected from {n_trials} out of"
            f" {n_total_trials} trials, with selection probability of"
            f" {selection_prob} and chance {chance}.  Final chance = {prob}.")
    return val


def get_probable_minimum_selected(
        n_total_trials: int, n_trials: int, selection_prob: float,
        chance: float = (1.0 / 100.0)) -> int:
    """
    Get the likely minimum number of items that will be selected from a
    set of `n_trials` from a total set of `n_total_trials`
    with a probability of selection of `selection_prob`.

    :returns: The likely minimum number of items
    """
    prob = (chance / float(n_total_trials))
    return binom.ppf(prob, n_trials, selection_prob)


def get_probability_within_range(distribution: RandomDistribution,
                                 lower: float, upper: float) -> float:
    """
    :param distribution:
    :param lower:
    :param upper:
    :returns: The probability that a value will fall within the given range
        for a given RandomDistribution.
    """
    stats = STATS_BY_NAME[distribution.name]
    return stats.cdf(distribution, upper) - stats.cdf(distribution, lower)


def get_maximum_probable_value(distribution: RandomDistribution, n_items: int,
                               chance: float = (1.0 / 100.0)) -> float:
    """
    :param distribution:
    :param n_items:
    :param chance:
    :returns: The likely maximum value of a RandomDistribution given a
        number of draws.
    """
    stats = STATS_BY_NAME[distribution.name]
    prob = 1.0 - (chance / float(n_items))
    return stats.ppf(distribution, prob)


def get_minimum_probable_value(distribution: RandomDistribution, n_items: int,
                               chance: float = (1.0 / 100.0)) -> float:
    """
    :param distribution: Random type to check
    :param n_items: Number of items to spread chance over
    :param chance: Possibility of a value
    :returns: The likely minimum value of a RandomDistribution given a
        number of draws.
    """
    stats = STATS_BY_NAME[distribution.name]
    prob = chance / float(n_items)
    return stats.ppf(distribution, prob)


def get_mean(distribution: RandomDistribution) -> float:
    """
    :param distribution:
    :returns: The mean of a RandomDistribution.
    """
    stats = STATS_BY_NAME[distribution.name]
    return stats.mean(distribution)


def get_standard_deviation(distribution: RandomDistribution) -> float:
    """
    :param distribution:
    :returns: The standard deviation of a RandomDistribution.
    """
    stats = STATS_BY_NAME[distribution.name]
    return stats.std(distribution)


def get_variance(distribution: RandomDistribution) -> float:
    """
    :param distribution:
    :returns: The variance of a RandomDistribution.
    """
    stats = STATS_BY_NAME[distribution.name]
    return stats.var(distribution)


def high(distribution: RandomDistribution) -> Optional[float]:
    """
    Gets the high or maximum boundary value for this distribution.

    Could return `None`.

    :param distribution:
    :returns: The maximum boundary value
    """
    stats = STATS_BY_NAME[distribution.name]
    return stats.high(distribution)


def low(distribution: RandomDistribution) -> Optional[float]:
    """
    Gets the low or minimum boundary value for this distribution.

    Could return `None`.

    :param distribution:
    :returns: The minimum boundary value
    """
    stats = STATS_BY_NAME[distribution.name]
    return stats.low(distribution)


def _validate_mars_kiss_64_seed(seed: List[int]) -> List[int]:
    """
    Update the seed to make it compatible with the RNG algorithm.
    """
    if seed[1] == 0:
        # y (<- seed[1]) can't be zero so set to arbitrary non-zero if so
        seed[1] = ARBITRARY_Y

    # avoid z=c=0 and make < 698769069
    seed[3] = seed[3] % MARS_C_MAX + 1
    return seed


def create_mars_kiss_seeds(rng: numpy.random.RandomState) -> Tuple[int, ...]:
    """
    Generates and checks that the seed values generated by the given
    random number generator or seed to a random number generator are
    suitable for use as a mars 64 kiss seed.

    :param rng: the random number generator.
    :return: a list of 4 integers which are used by the mars64 kiss random
        number generator for seeds.
    """
    kiss_seed = _validate_mars_kiss_64_seed([
        rng.randint(-BASE_RANDOM_FOR_MARS_64, CAP_RANDOM_FOR_MARS_64) +
        BASE_RANDOM_FOR_MARS_64 for _ in range(N_RANDOM_NUMBERS)])
    return tuple(kiss_seed)


def get_n_bits(n_values: int) -> int:
    """
    Determine how many bits are required for the given number of values.

    :param n_values: the number of values (starting at 0)
    :return: the number of bits required to express that many values
    """
    if n_values == 0:
        return 0
    if n_values == 1:
        return 1
    return int(math.ceil(math.log2(n_values)))


def get_time_to_write_us(n_bytes: int, n_cores: int) -> int:
    """
    Determine how long a write of a given number of bytes will take in us.

    No known use so may be removed.

    :param n_bytes: The number of bytes to transfer
    :param n_cores: How many cores will be writing at the same time
    :returns: how long a write of a given number of bytes will take in us.
    """
    bandwidth_per_core = WRITE_BANDWIDTH_BYTES_PER_SECOND / n_cores
    seconds = n_bytes / bandwidth_per_core
    return int(math.ceil(seconds * MICRO_TO_SECOND_CONVERSION))


def get_neo_io(file_or_folder: str) -> BaseIO:
    """
    Hack for https://github.com/NeuralEnsemble/python-neo/issues/1287

    In Neo 0.12 neo.get_io only works with existing files

    :param file_or_folder:
    :returns: Neo IO instance, guessing the type based on the filename suffix.
    """
    try:
        return neo.get_io(file_or_folder)
    except ValueError as ex:
        try:
            _, suffix = os.path.splitext(file_or_folder)
            suffix = suffix[1:].lower()
            if suffix in neo.io_by_extension:
                writer_list = neo.io_by_extension[suffix]
                return writer_list[0](file_or_folder)
        except AttributeError:
            # for older neo which has no io_by_extension
            pass
        raise ex


def report_non_spynnaker_pynn(msg: str) -> None:
    """
    Report a case of non-spynnaker-compatible PyNN being used.  This will warn
    or error depending on the configuration setting.

    :param msg: The message to report
    """
    if get_config_bool("Simulation", "error_on_non_spynnaker_pynn"):
        raise ConfigurationException(msg)
    else:
        warn_once(logger, msg)


def check_rng(rng: AbstractRNG, where: str) -> None:
    """
    Check for non-None rng parameter since this is no longer compatible with
    sPyNNaker.  If not None, warn or error depending on a config value.

    :param rng: The rng parameter value.
    :param where: Caller location for error message
    """
    if rng is not None and rng.seed is not None:
        report_non_spynnaker_pynn(
            f"Use of rng in {where} is not supported in sPyNNaker in this"
            " case. Please instead use seed=<seed> in the target Population to"
            " ensure random numbers are seeded.")



# Copyright (c) 2014 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .abstract_timing_dependence import AbstractTimingDependence
from .timing_dependence_spike_pair import TimingDependenceSpikePair
from .timing_dependence_pfister_spike_triplet import (
    TimingDependencePfisterSpikeTriplet)
from .timing_dependence_recurrent import TimingDependenceRecurrent
from .timing_dependence_spike_nearest_pair import (
    TimingDependenceSpikeNearestPair)
from .timing_dependence_vogels_2011 import TimingDependenceVogels2011

__all__ = [
    "AbstractTimingDependence", "TimingDependenceSpikePair",
    "TimingDependencePfisterSpikeTriplet", "TimingDependenceRecurrent",
    "TimingDependenceSpikeNearestPair", "TimingDependenceVogels2011"
]
